'''
Experiment constructors. 'Experiment' instances are the combination of
a task and a list of features.  Rather than have a separate class for
all the possible combinations of tasks and features, a custom class for
the experiment is created programmatically using 'type'. The created class
has methods of the base task as well as all the selected features.
'''
import numpy as np

##################
##### Traits #####
##################
try:
    import traits.api as traits
except ImportError:
    import enthought.traits.api as traits


class InstanceFromDB(traits.Instance):
    def __init__(self, *args, **kwargs):
        if 'bmi3d_db_model' in kwargs:
            self.bmi3d_db_model = kwargs['bmi3d_db_model']
        else:
            raise ValueError("If using trait 'InstanceFromDB', must specify bmi3d_db_model!")

        # save the arguments for the database
        #self.bmi3d_query_kwargs = kwargs.pop('bmi3d_query_kwargs', dict())
        if 'bmi3d_query_kwargs' in kwargs:
            self.bmi3d_query_kwargs = kwargs['bmi3d_query_kwargs']
        else:
            self.bmi3d_query_kwargs = dict()

        super(InstanceFromDB, self).__init__(*args, **kwargs)


class DataFile(InstanceFromDB):
    def __init__(self, *args, **kwargs):
        kwargs['bmi3d_db_model'] = 'DataFile'
        super(DataFile, self).__init__(*args, **kwargs)


class OptionsList(traits.Enum):
    '''
    Wrapper around Enum so that we can keep track of the possible enumerations in list
    called 'bmi3d_input_options' which will be hidden in the UI
    '''
    def __init__(self, *args, **kwargs):
        if 'bmi3d_input_options' not in kwargs:
            kwargs['bmi3d_input_options'] = args[0]
        super(OptionsList, self).__init__(*args, **kwargs)


traits.InstanceFromDB = InstanceFromDB
traits.DataFile = DataFile
traits.OptionsList = OptionsList



from . import experiment
from . import generate
from . import report
from .experiment import Experiment, LogExperiment, Sequence, TrialTypes, FSMTable, StateTransitions

from . import task_wrapper

try:
    from .Pygame import Pygame
except:
    import warnings
    warnings.warn('riglib/experiment/__init__.py: could not import Pygame (note capital P)')
    Pygame = object

def make(exp_class, feats=(), verbose=False):
    '''
    Creates a class which inherits from a base experiment class as well as a set of optional features.
    This function is a *metafunction* as it returns a custom class construction.

    Parameters
    ----------
    exp_class : class
        Base class containing the finite state machine of the task
    feats : iterable of classes
        Additional classes from which to also inherit

    Returns
    -------
    class
        New class which inherits from the base 'exp_class' and the selected 'feats'
    '''
    if len(feats) == 0:
        return exp_class
    else:
        # construct the class list to define inheritance order for the custom task
        # inherit from the features first, then the base class
        clslist = tuple(feats) + (exp_class,)

        if verbose:
            print("metaclass constructor", clslist, feats)

        # return custom class
        return type(exp_class.__name__, clslist, dict())




# Generated by Django 3.1.4 on 2021-10-17 15:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0010_experimenter'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='experimenter',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='tracker.experimenter'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='taskentry',
            name='project',
            field=models.TextField(default='testing'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='taskentry',
            name='session',
            field=models.TextField(default='testing'),
            preserve_default=False,
        ),
    ]



import numpy as np
from dataclasses import dataclass, field
from typing import Any, Dict, NamedTuple, Optional, Tuple, Union

import cv2

@dataclass
class NoDelaySurfaceGazeMapper:
    camera: "RadialDistortionCamera"
    homography: Optional[np.ndarray] = field(default=None)

    def update_homography(self, homography: np.ndarray):
        homography = np.array(homography)
        if homography.shape != (3, 3):
            raise ValueError(
                f"Input has incorrect shape! Expected 3x3, got {homography}"
            )
        self.homography = homography

    def gaze_to_surface(
        self, norm_pos: Tuple[float, float]
    ) -> Optional["SurfaceMappedGaze"]:
        if self.homography is None:
            return None

        x, y = self.denormalize(*norm_pos)
        points = self.camera.undistort_points_on_image_plane([x, y])
        points.shape = -1, 1, 2
        points = cv2.perspectiveTransform(points, self.homography)
        points.shape = -1
        return SurfaceMappedGaze(*points)

    def denormalize(self, x: float, y: float) -> Tuple[float, float]:
        return self.camera.resolution[0] * x, (1 - y) * self.camera.resolution[1]


@dataclass
class SurfaceMappedGaze:
    norm_x: float
    norm_y: float
    on_surface: bool = field(init=False)
    timestamp: Optional[float] = field(default=None)
    confidence: Optional[float] = field(default=None)

    def __post_init__(self):
        self.on_surface = (0.0 <= self.norm_x <= 1.0) and (0.0 <= self.norm_y <= 1.0)


@dataclass
class RadialDistortionCamera:
    """Camera model assuming a lense with radial distortion."""

    resolution: Tuple[int, int]
    cam_matrix: np.ndarray #npt.ArrayLike
    dist_coefs: np.ndarray #npt.ArrayLike

    def __post_init__(self):
        self.cam_matrix = np.array(self.cam_matrix)
        self.dist_coefs = np.array(self.dist_coefs)

    # CameraModel Interface
    def undistort_points_on_image_plane(self, points):
        points = self.__unprojectPoints(points, use_distortion=True)
        points = self.__projectPoints(points, use_distortion=False)
        return points

    # Private
    def __projectPoints(self, object_points, rvec=None, tvec=None, use_distortion=True):
        """
        Projects a set of points onto the camera plane as defined by the camera model.
        :param object_points: Set of 3D world points
        :param rvec: Set of vectors describing the rotation of the camera when recording the corresponding object point
        :param tvec: Set of vectors describing the translation of the camera when recording the corresponding object point
        :return: Projected 2D points
        """
        input_dim = object_points.ndim
 # # matrix for camera distortion
        # camera = RadialDistortionCamera(
        #     resolution=(1280, 720),
        #     cam_matrix=[
        #         [794.3311439869655, 0.0, 633.0104437728625],
        #         [0.0, 793.5290139393004, 397.36927353414865],
        #         [0.0, 0.0, 1.0],
        #     ],
        #     dist_coefs=[
        #         [
        #             -0.3758628065070806,
        #             0.1643326166951343,
        #             0.00012182540692089567,
        #             0.00013422608638039466,
        #             0.03343691733865076,
        #             0.08235235770849726,
        #             -0.08225804883227375,
        #             0.14463365333602152,
        #         ]
        #     ],
        # )

        # self.mapper = NoDelaySurfaceGazeMapper(camera)
        # self.mapped_points = []
        object_points = object_points.reshape((1, -1, 3))

        if rvec is None:
            rvec = np.zeros(3).reshape(1, 1, 3)
        else:
            rvec = np.array(rvec).reshape(1, 1, 3)

        if tvec is None:
            tvec = np.zeros(3).reshape(1, 1, 3)
        else:
            tvec = np.array(tvec).reshape(1, 1, 3)

        if use_distortion:
            _D = self.dist_coefs
        else:
            _D = np.asarray([[0.0, 0.0, 0.0, 0.0, 0.0]])

        image_points, jacobian = cv2.projectPoints(
            object_points, rvec, tvec, self.cam_matrix, _D
        )

        if input_dim == 2:
            image_points.shape = (-1, 2)
        elif input_dim == 3:
            image_points.shape = (-1, 1, 2)
        return image_points

    def __unprojectPoints(self, pts_2d, use_distortion=True, normalize=False):
        """
        Undistorts points according to the camera model.
        :param pts_2d, shape: Nx2
        :return: Array of unprojected 3d points, shape: Nx3
        """
        pts_2d = np.array(pts_2d, dtype=np.float32)

        # Delete any posibly wrong 3rd dimension
        if pts_2d.ndim == 1 or pts_2d.ndim == 3:
            pts_2d = pts_2d.reshape((-1, 2))

        # Add third dimension the way cv2 wants it
        if pts_2d.ndim == 2:
            pts_2d = pts_2d.reshape((-1, 1, 2))

        if use_distortion:
            _D = self.dist_coefs
        else:
            _D = np.asarray([[0.0, 0.0, 0.0, 0.0, 0.0]])

        pts_2d_undist = cv2.undistortPoints(pts_2d, self.cam_matrix, _D)

        pts_3d = cv2.convertPointsToHomogeneous(pts_2d_undist)
        pts_3d.shape = -1, 3

        if normalize:
            pts_3d /= np.linalg.norm(pts_3d, axis=1)[:, np.newaxis]

        return pts_3d


'''
Features for interacting with the white-matter ecube
'''
import time
import os
import numpy as np
from datetime import datetime
from riglib.experiment import traits
from riglib import ecube, source
from features.neural_sys_features import CorticalBMI, CorticalData
import traceback

ecube_path = "/media/NeuroAcq" # TODO this should be configurable elsewhere
log_path = os.path.join(os.path.dirname(__file__), '../log')
log_filename = os.path.join(log_path, "ecube_log")

def make_ecube_session_name(saveid):
    now = datetime.now()
    session = now.strftime("%Y-%m-%d")
    number = str(saveid) if saveid else "Test"
    return session + "_BMI3D_te" + number

def log_str(s, mode="a", newline=True):
    if newline and not s.endswith("\n"):
        s += "\n"
    with open(log_filename, mode) as fp:
        fp.write(s)

def log_exception(err, mode="a"):
    with open(log_filename, mode) as fp:
        traceback.print_exc(file=fp)
        fp.write(str(err))

class RecordECube(traits.HasTraits):

    record_headstage = traits.Bool(False, desc="Should we record headstage data?")
    headstage_connector = traits.Int(7, desc="Which headstage input to record (1-indexed)")
    headstage_channels = traits.Tuple((1, 1), desc="Range of headstage channels to record (1-indexed)")
    headstage_module_id = traits.String("", desc="Unique identifier for the tether module being used")
    headstage_stacking_id = traits.String("", desc="Unique identifier for the stacking cable being used")
    channel_mapping_file = traits.String("", desc="Name of channel mapping excel file")
    drmap_chamber = traits.String("", desc="Name of the recording chamber (e.g. LM1)")
    drmap_drive_type = traits.String("", desc="Type of recording drive (e.g. ECoG244)")
    drmap_drive_id = traits.String("", desc="Unique identifier for the drive being used")
    drmap_drive_orientation = traits.String("", desc="Orientation (in degrees) of the drive inside the chamber")
    drmap_implant_date = traits.String("", desc="Date recording drive was implanted (YYMMDD)")
    drmap_config_date = traits.String("", desc="Date channel mapping was configured (YYMMDD)")
    ecube_feature_version = traits.Int(1, desc="Version number of the BMI3D feature used to record ecube data")
    
    ecube_status = "Not initialized"
    hidden_traits = ['ecube_feature_version']

    def cleanup(self, database, saveid, **kwargs):
        '''
        Function to run at 'cleanup' time, after the FSM has finished executing. See riglib.experiment.Experiment.cleanup
        This 'cleanup' method remotely stops the plexon file recording and then links the file created to the database ID for the current TaskEntry
        '''
        super_result = super().cleanup(database, saveid, **kwargs)
        
        # Check that we actually started recording
        if not self.ecube_status == "recording":
            return super_result
        
        ecube_session = make_ecube_session_name(saveid) # usually correct, but might be wrong if running overnight!

        # Stop recording
        try:
            self._ecube_cleanup(saveid)
        except Exception as e:
            print(e)
            traceback.print_exc()
            log_exception(e)
            print('\n\neCube recording could not be stopped! Please manually stop the recording\n\n')
            # return False
            # Actually we should still save the recording to the database, since no files have to be moved. LS 9/10/2021

        # Save file to database
        print("Saving ecube file to database...")
        dbname = kwargs['dbname'] if 'dbname' in kwargs else 'default'
        suffix = '' # note: database functions don't take keyword arguements like custom_suffix=suffix
        filepath = os.path.join(ecube_path, ecube_session)
        if dbname == 'default':
            database.save_data(filepath, "ecube", saveid, False, False, suffix) # make sure to add the ecube datasource!
        else:
            database.save_data(filepath, "ecube", saveid, False, False, suffix, dbname=dbname)
        return super_result

    def _ecube_cleanup(self, saveid):
        '''
        Cleanup function that gets run at the end of the experiment regardless of success.
        '''
        from riglib.ecube import utils, pyeCubeStream

        # Mark the end of the recording
        if hasattr(self, 'sync_code'):
            print('Sending stop recording pulse')
            self.sync_code(1 << self.sync_params['recording_pin'])
            if hasattr(self, 'pulse'):
                self.pulse.join()

        else:
            print('No sync feature enabled. Cannot send stop pulse.')

        # Stop the recording
        ec = pyeCubeStream.eCubeStream()
        active = ec.listremotesessions()
        stopped = saveid is None
        for session in active:
            if str(saveid) in session:
                ecube_session = session
                time.sleep(1) # Need to wait for a bit since the recording system has some latency and we don't want to stop prematurely
                ec.remotestopsave(session)
                stopped = True

        # Remove headstage sources so they can be added again later
        utils.remove_headstage_sources(ec)

        # Check that everything worked out
        if stopped:
            print('Stopped eCube recording session ' + session)
            log_str("Stopped recording " + session)
        else:
            raise Exception('Could not find active session for ' + ecube_session)

    
    @classmethod
    def pre_init(cls, saveid=None, record_headstage=False, headstage_connector=None, headstage_channels=None, **kwargs):
        '''
        Run prior to starting the task to remotely start recording from the plexon system
        '''
        cls.ecube_status = "Not initialized"
        if saveid is not None:
            from riglib.ecube import pyeCubeStream
            session_name = make_ecube_session_name(saveid)
            log_str("\n\nNew recording for task entry {}: {}".format(saveid, session_name))
            try:
                ec = pyeCubeStream.eCubeStream()
                ec.add(('AnalogPanel',))
                ec.add(('DigitalPanel',))
                if record_headstage:
                    ec.add(('Headstages', headstage_connector, headstage_channels))
                ec.remotesave(session_name)
                if record_headstage:
                    ec.remove(('Headstages', headstage_connector))
                ec.remove(('AnalogPanel',))
                ec.remove(('DigitalPanel',))
                active_sessions = ec.listremotesessions()
                if session_name in active_sessions:
                    cls.ecube_status = "recording"
                else:
                    cls.ecube_status = "Could not start eCube recording. Make sure servernode is running!\n"
            except Exception as e:
                print(e)
                traceback.print_exc()
                log_exception(e)
                cls.ecube_status = e
            
        else:
            # Just a test, try to connect to servernode to make sure it's working
            from riglib.ecube import pyeCubeStream
            session_name = make_ecube_session_name(saveid)
            log_str("\n\nNew recording for task entry {}: {}".format(saveid, session_name))
            try:
                ec = pyeCubeStream.eCubeStream()
                ec.add(('AnalogPanel',))
                ec.add(('DigitalPanel',))
                if record_headstage:
                    ec.add(('Headstages', headstage_connector, headstage_channels))
                    ec.remove(('Headstages', headstage_connector))
                ec.remove(('AnalogPanel',))
                ec.remove(('DigitalPanel',))
                cls.ecube_status = "testing"
            except Exception as e:
                print(e)
                traceback.print_exc()
                log_exception(e)
                cls.ecube_status = "Could not connect to eCube. Make sure servernode is running!\n"
        if hasattr(super(), 'pre_init'):
            super().pre_init(saveid=saveid, **kwargs)

    def run(self):
        if not self.ecube_status in ["testing", "recording"]:
            import io
            self.terminated_in_error = True
            self.termination_err = io.StringIO()
            self.termination_err.write(self.ecube_status)
            self.termination_err.seek(0)
            self.state = None
        try:
            # Send a pulse to mark the beginning of the recording
            time.sleep(0.1) # Wait a bit to be sure the recording started
            if hasattr(self, 'sync_code'):
                print('Sending start recording pulse')
                self.sync_code(1 << self.sync_params['recording_pin'])
                if hasattr(self, 'pulse'):
                    self.pulse.join()
            else:
                print('No sync feature enabled. Cannot send start pulse.')
            time.sleep(0.1)
            super().run()
        except Exception as e:
            try:
                self._ecube_cleanup(self.saveid)
            except Exception as ee:
                log_str("Tried to cleanup ecube recording but couldn't...")
                log_exception(ee)
            finally:
                raise e

class EcubeFileData(CorticalData, traits.HasTraits):
    '''
    Streams data from an ecube file into BMI3D as neurondata
    '''

    ecube_bmi_filename = traits.String("", desc="File to playback in BMI")
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Query the file to see the channels
        file = ecube.file.parse_file(self.ecube_bmi_filename)
        self.cortical_channels = file.channels

        # These get read by CorticalData when initializing the extractor
        self._neural_src_type = source.MultiChanDataSource
        self._neural_src_kwargs = dict(
            send_data_to_sink_manager=self.send_data_to_sink_manager, 
            channels=self.cortical_channels,
            ecube_bmi_filename=self.ecube_bmi_filename)
        self._neural_src_system_type = ecube.File
        
    @property 
    def sys_module(self):
        return ecube   

class ManualPositionDecodeTest():
    '''
    Test feature for ecube data
    '''
    
    def init(self):
        super().init()
        from riglib.bmi.extractor import LFPMTMPowerExtractor
        self.extractor = LFPMTMPowerExtractor(self.neurondata, channels=[1], bands=[(50,200)])

    def _get_manual_position(self):
        '''
        Fetches neurondata position
        '''
        if not hasattr(self, 'neurondata'):
            return
        ext = self.extractor(4) # start time can be anything
        
        power = ext['lfp_power'][0][0]
        pt = [-power, 0, 0]

        print(pt)
        return [pt]

class EcubeData(CorticalData, traits.HasTraits):
    '''
    Streams neural data using ecube as the datasource. Use this if you want online neural data
    but don't want to use a decoder.
    '''

    streaming_channels = traits.Tuple((1,1), desc="Range of channels to stream")

    def init(self):
        self.cortical_channels = list(range(*self.streaming_channels))
        super().init()

    @property 
    def sys_module(self):
        return ecube   

class EcubeFileBMI(EcubeFileData, CorticalBMI):
    '''
    Streams data from an ecube file into a BMI decoder
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cortical_channels = [int(ch) for ch in self.decoder.units[:,0]]

        # These get read by CorticalData when initializing the extractor
        self._neural_src_type = source.MultiChanDataSource
        self._neural_src_kwargs = dict(
            send_data_to_sink_manager=self.send_data_to_sink_manager, 
            channels=self.cortical_channels,
            ecube_bmi_filename=self.ecube_bmi_filename)
        self._neural_src_system_type = ecube.File

class EcubeBMI(CorticalBMI):
    '''
    BMI using ecube as the datasource.
    '''

    bmi_ecube_headstage = traits.Int(7, desc="Which headstage to use for BMI data")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._neural_src_system_type = ecube.LFP

    def init(self):
        self.neural_src_kwargs = dict(headstage=self.bmi_ecube_headstage)
        super().init()

    @property 
    def sys_module(self):
        return ecube   


class TestExperiment():
    state = None
    def __init__(self, *args, **kwargs):
        pass

    def pre_init(saveid, **kwargs):
        pass

    def start(self):
        import time
        time.sleep(1)

    def cleanup(self, database, saveid, **kwargs):
        pass

    def terminate(self):
        pass

    def join(self):
        pass

class TestClass(RecordECube, TestExperiment):
    pass

if __name__ == "__main__":

    TestClass.pre_init(saveid=10)
    exp = TestClass()
    exp.start()
    exp.cleanup(None, 10)
    # from riglib import experiment
    # proc = experiment.task_wrapper.TaskWrapper(
    #     subj=None, params=dict(), target_class=TestClass, saveid=10)
    # proc.start()


'''
Compile a list of all the files which should be backed up
'''
import os
from . import dbfunctions as dbfn
from .tracker import models
import datetime

#Get backup list:
backed_up_tes = list(models.TaskEntry.objects.filter(backup=True, date__gte=datetime.datetime(2015, 11, 0o1)))

#Get list of Task Entries already added to list: 
try:
    te_added = open(os.path.expandvars('$HOME/files_to_backup_te_list'), 'r')
    items = te_added.read().split('\n')
    already_te_added = [int(i) for i in items if len(i) > 0]
    te_added.close()
except:
    print('No previous backup list found, making new one')
    te_added = open(os.path.expandvars('$HOME/files_to_backup_te_list'), 'w')
    te_added.close()
    already_te_added = []

systems_to_back_up = ['hdf', 'plexon']

#Open file for writing 
#Open new file: 
f = open(os.path.expandvars('$HOME/files_to_backup'), 'w')

#Append to previous file: 
f_list = open(os.path.expandvars('$HOME/files_to_backup_te_list'), 'a')

for k,te in enumerate(backed_up_tes):

    #Check if already backed up: 
    if te.id in already_te_added:
        pass
    else:
        #Array for datafile names:
        datafiles = []
        try:
            te = dbfn.TaskEntry(te)
            te_datafiles = te.datafiles
            for sysname in systems_to_back_up:
                if sysname in te_datafiles:
                    datafiles.append(te_datafiles[sysname])

            #Close HDF file to free memory: 
            try:
                te.hdf.close()
            except:
                pass


            if k % 100 == 0:
                print(k)

            if not te.decoder_record is None:
                datafiles.append(te.decoder_filename)
        except:
            pass

        #Write to file: 
        if len(datafiles) > 0:
            for datafile in datafiles:
                rel_datafile = os.path.relpath(datafile, '/storage')
                f.write("%s\n" % rel_datafile)
        #Write to 'already backed up file:'
        f_list.write("%s\n" % str(te.id))

f.close()
f_list.close()

# # don't back up video files
# datafiles = filter(lambda x: not (x[-4:] == '.avi'), datafiles)

# for datafile in datafiles:
#     rel_datafile = os.path.relpath(datafile, '/storage')
#     f.write("%s\n" % rel_datafile)
# f.close()



"""
Robot object.

Python implementation by: Luis Fernando Lara Tobar and Peter Corke.
Based on original Robotics Toolbox for Matlab code by Peter Corke.
Permission to use and copy is granted provided that acknowledgement of
the authors is made.

@author: Luis Fernando Lara Tobar and Peter Corke
"""

from numpy import *
from .utility import *
from .transform import *
import copy
from .Link import *

class Robot(object):
    """Robot object.
    Instances of this class represent a robot manipulator
    within the toolbox.
    """
        
    def __init__(self, arg=None, gravity=None, base=None, tool=None, name='', comment='', manuf=''):
        """
        Robot object constructor.  Create a robot from a sequence of Link objects.
        
        Several basic forms exist:
            - Robot()        create a null robot
            - Robot(robot)   create a clone of the robot object
            - Robot(links)   create a robot based on the passed links
            
        Various options can be set using named arguments:
        
            - gravity; gravitational acceleration (default=[0,0,9.81])
            - base; base transform (default 0)
            - tool; tool transform (default 0)
            - name
            - comment
            - manuf
        """

        if isinstance(arg, Robot):
            for k,v in arg.__dict__.items():
                if k == "links":
                    self.__dict__[k] = copy.copy(v);           
                else:
                    self.__dict__[k] = v;           
        elif len(arg) > 1 and isinstance(arg[0], Link):
            self.links = arg;
        else:
            raise AttributeError;

        # fill in default base and gravity direction
        if gravity != None:
            self.gravity = gravity;
        else:
            self.gravity = [0, 0, 9.81];
        
        if base != None:
            self.base = base;
        else:
            self.base = mat(eye(4,4));
        
        if tool != None:
            self.tool = tool;
        else:
            self.tool = mat(eye(4,4));

        if manuf:
            self.manuf = manuf
        if comment:
            self.comment = comment
        if name:
            self.name = name

        #self.handle = [];
        #self.q = [];
        #self.plotopt = {};
        #self.lineopt = {'Color', 'black', 'Linewidth', 4};
        #self.shadowopt = {'Color', 'black', 'Linewidth', 1};

        return None;

    def __str__(self):
        s = 'ROBOT(%s, %s)' % (self.name, self.config());
        return s;
        
    def __repr__(self):
        s = '';
        if self.name:
            s += 'name: %s\n' % (self.name)
        if self.manuf:
            s += 'manufacturer: %s\n' % (self.manuf)
        if self.comment:
            s += 'commment: %s\n' % (self.comment)
        
        for link in self.links:
            s += str(link) + '\n';
        return s;   

    def __mul__(self, r2):
        r = Robot(self);        # clone the robot
        r.links += r2.links;
        return r;

    def copy(self):
        """
        Return a copy of the Robot object
        """
        return copy.copy(self);
               
    def ismdh(self):
        return self.mdh;
        
    def config(self):
        """
        Return a configuration string, one character per joint, which is
        either R for a revolute joint or P for a prismatic joint.
        For the Puma560 the string is 'RRRRRR', for the Stanford arm it is 'RRPRRR'.
        """
        s = '';
        
        for link in self.links:
            if link.sigma == 0:
                s += 'R';
            else:
                s += 'P';
        return s;

    def nofriction(self, all=False):
        """
        Return a Robot object where all friction parameters are zero.
        Useful to speed up the performance of forward dynamics calculations.
        
        @type all: boolean
        @param all: if True then also zero viscous friction
        @see: L{Link.nofriction}
        """
        r = Robot(self);
        r.name += "-nf";
        newlinks = [];
        for oldlink in self.links:
            newlinks.append( oldlink.nofriction(all) );
        r.links = newlinks;
        return r;

    def __setattr__(self, name, value):
        """
        Set attributes of the robot object
        
            - robot.name = string (name of this robot)
            - robot.comment = string (user comment)
            - robot.manuf = string (who built it)
            - robot.tool = 4x4 homogeneous tranform
            - robot.base = 4x4 homogeneous tranform
            - robot.gravity = 3-vector  (gx,gy,gz)
        """
        
        if name in ["manuf", "name", "comment"]:
            if not isinstance(value, str):
                raise ValueError('must be a string')
            self.__dict__[name] = value;
            
        elif name == "links":
            if not isinstance(value[0], Link):
                raise ValueError('not a Link object')
            self.__dict__[name] = value;
            self.__dict__['n'] = len(value);
            # set the robot object mdh status flag
            for link in self.links:
                if link.convention != self.links[0].convention:
                    raise 'robot has mixed D&H link conventions'
            self.__dict__['mdh'] = self.links[0].convention == Link.LINK_MDH;
            
        elif name == "tool":
            if not ishomog(value):
                raise ValueError('tool must be a homogeneous transform')
            self.__dict__[name] = value;

        elif name == "gravity":
            v = arg2array(value);
            if len(v) != 3:
                raise ValueError('gravity must be a 3-vector')
            self.__dict__[name] = mat(v).T
            
        elif name == "base":
            if not ishomog(value):
                raise ValueError('base must be a homogeneous transform')
            self.__dict__[name] = value;
            
        else:
            raise AttributeError;





import numpy as np
from ..bmi import clda
from ..bmi import train

kinarm_bands = []
for i in np.arange(0,100,10):
    kinarm_bands.extend([[i, i+10]])
kinarm_bands.extend([[25, 40],[40, 55], [65, 90], [2, 100]])

class StateHolder(object):
    def __init__(self, x_mat, A_mat, powercap_flag, zbound, *args, **kwargs):
        if powercap_flag:
            pos_mean = np.sum((np.multiply(x_mat, A_mat)*0)+zbound[0], axis=0)
            self.mean = np.squeeze(np.hstack((np.array([pos_mean]), np.zeros((1,4)))))
        else:
            #self.mean = np.dot(x_array, A_array)
            pos_mean = np.sum(np.multiply(x_mat, A_mat), axis=0)
            self.mean = np.squeeze(np.hstack((np.array([pos_mean]), np.zeros((1,4)))))

class SmoothFilter(object):
    '''Moving Avergae Filter used in 1D or 2D LFP control:
    x_{t} = a0*x_{t} + a1*x_{t-1} + a2*x_{t-2} + ...

    Parameters

    ----------
    A: np.array of shape (N, 3)
        Weights for previous states
    X: np. array of previous states (N, 3)
    '''

    def __init__(self, n_steps, **kwargs):
        self.n_steps = n_steps
        A = np.ones(( self.n_steps, ))/float(self.n_steps)
        self.A = np.tile(np.array([A]).T, [1,3])

        self.control_method = 'fraction'
        self.current_lfp_pos = 0
        self.current_powercap_flag = 0

    def get_mean(self):
        return np.array(self.state.mean).ravel()

    def init_from_task(self,**kwargs):
        if 'n_steps' in kwargs:
            self.n_steps = kwargs['n_steps']
            A = np.ones(( self.n_steps, ))/float(self.n_steps)
            self.A = np.tile(np.array([A]).T, [1,3])

        if 'powercap' in kwargs:
            self.powercap = kwargs['powercap']

        if 'zboundaries' in kwargs:
            self.zboundaries = kwargs['zboundaries']

        if 'lfp_frac_lims' in kwargs:
            self.frac_lims = kwargs['lfp_frac_lims']

        if 'xlfp_frac_lims' in kwargs:
            self.xfrac_lims = kwargs['xlfp_frac_lims']

    def _init_state(self, init_state=None,**kwargs):
        if init_state is None:
            self.X = np.zeros(( self.n_steps, 3))

        #Implemented later: 
        elif init_state is 'average':
            if self.control_method == 'fraction':
                mn = np.mean(np.array(kwargs['frac_lim']))
            elif self.control_method == 'power':
                mn = np.mean(np.array(kwargs['pwr_mean']))
            self.X = np.zeros(( self.n_steps, 3 )) + mn

        self.state = StateHolder(self.X, self.A, 0, 0)
        
        self.cnt = 0

    def __call__(self, obs, **kwargs):
        self.state = self._mov_avg(obs, **kwargs)

    def _mov_avg(self, obs,**kwargs):
        #self.zboundaries = kwargs['zboundaries']
        self.fft_inds = kwargs['fft_inds']
        obs = obs.reshape(len(kwargs['channels']), len(kwargs['fft_freqs']))
        
        self.current_lfp_pos, self.current_powercap_flag = self.get_lfp_cursor(obs)
        
        self.X = np.vstack(( self.X[1:,:], self.current_lfp_pos ))
        return StateHolder(self.X, self.A, self.current_powercap_flag, self.zboundaries)

    def get_lfp_cursor(self, psd_est):
        'Obs: channels x frequencies '
        # Control band: 
        c_idx = self.control_band_ind

        #As done in kinarm script, sum together frequencies within a band, then take the mean across channels
        c_val = np.mean(np.sum(psd_est[:, self.fft_inds[c_idx]], axis=1))

        xc_idx = self.x_control_band_ind
        xc_val = np.mean(np.sum(psd_est[:, self.fft_inds[xc_idx]], axis=1))

        p_idx = self.totalpw_band_ind
        p_val = np.mean(np.sum(psd_est[:, self.fft_inds[p_idx]], axis=1))

        if self.control_method == 'fraction':
            lfp_control = c_val / float(p_val)
            xlfp_control = xc_val/float(p_val)

        elif self.control_method == 'power':
            lfp_control = c_val

        cursor_pos = self.lfp_to_cursor(lfp_control, xlfp_control)

        #if p_val <= self.powercap:
        #write c_val, xc_val, p_val, to file:
        if self.cnt < 3000:
            self.files[0].write(str(c_val)+',')
            self.files[1].write(str(p_val)+',')
            self.files[2].write(str(xc_val)+',')
            self.cnt += 1
        elif self.cnt == 3000:
            for f in self.files:
                f.close()
                
        #Hack: make x axis control powercap value:
        if xc_val <= self.powercap:
            powercap_flag = 0
        else:
            powercap_flag = 1

        return cursor_pos, powercap_flag

    def lfp_to_cursor(self, lfppos, xlfp_control):
        if self.control_method == 'fraction':
            dmn = lfppos - np.mean(self.frac_lims);
            cursor_pos = dmn * (self.zboundaries[1]-self.zboundaries[0]) / (self.frac_lims[1] - self.frac_lims[0])
            
            #Xcursor postion, keep within (0, -16) boundaries: 
            xdmn = xlfp_control - np.mean(self.xfrac_lims)
            xcursor_pos = xdmn * (0--16) / (self.xfrac_lims[1] - self.xfrac_lims[0])
            xcursor_pos = xcursor_pos - 8; #nonzero offset
            return np.array([xcursor_pos, 0, cursor_pos])


    def _pickle_init(self):
        pass

from .bmi import Decoder
class One_Dim_LFP_Decoder(Decoder):

    def __init__(self, *args, **kwargs):
        
        #Args: sf, units, ssm, extractor_cls, extractor_kwargs
        super(One_Dim_LFP_Decoder, self).__init__(args[0], args[1], args[2])
        
        self.extractor_cls = args[3]
        self.extractor_kwargs = args[4]

        #For now, hardcoded: 
        bands = kinarm_bands
        control_method='fraction'
        no_log=True
        no_mean=True

        self.extractor_kwargs['bands'] = bands
        self.extractor_kwargs['no_log'] = no_log
        self.extractor_kwargs['no_mean'] = no_mean

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self,key,value)

    def predict(self, neural_obs, **kwargs):
        #kwargs['zboundaries'] = self.filt.zboundaries
        kwargs['fft_inds'] = self.extractor_kwargs['fft_inds']
        kwargs['channels'] = self.extractor_kwargs['channels']
        kwargs['fft_freqs'] = self.extractor_kwargs['fft_freqs']

        self.filt(neural_obs, **kwargs)


    def init_from_task(self,**kwargs):
        if 'lfp_control_band' in kwargs:
            self.filt.control_band_ind, self.extractor_kwargs['bands'], self.extractor_kwargs['fft_inds'] = \
            self._get_band_ind(self.extractor_kwargs['fft_freqs'], kwargs['lfp_control_band'], self.extractor_kwargs['bands'])


        if 'lfp_totalpw_band' in kwargs:
            self.filt.totalpw_band_ind, self.extractor_kwargs['bands'], self.extractor_kwargs['fft_inds'] = \
            self._get_band_ind(self.extractor_kwargs['fft_freqs'], kwargs['lfp_totalpw_band'], self.extractor_kwargs['bands'])

        if 'xlfp_control_band' in kwargs:
            self.filt.x_control_band_ind, self.extractor_kwargs['bands'], self.extractor_kwargs['fft_inds'] = \
            self._get_band_ind(self.extractor_kwargs['fft_freqs'], kwargs['xlfp_control_band'], self.extractor_kwargs['bands'])

        c_txt = open('/home/helene/Downloads/pk/txt_write/control.txt','w')
        p_txt = open('/home/helene/Downloads/pk/txt_write/tot.txt','w')
        xc_txt = open('/home/helene/Downloads/pk/txt_write/x_cont.txt','w')
        self.filt.files = [c_txt, p_txt, xc_txt]

    def _get_band_ind(self, freq_pts, band, band_set):
        band_ind = -1
        for b, bd in enumerate(band_set):
            if (bd[0]==band[0]) and (bd[1]==band[1]):
                band_ind = b
        if band_ind == -1:
            band_ind = b+1
            band_set.extend([band])

        fft_ind = dict()
        for band_idx, band in enumerate(band_set):
            fft_ind[band_idx] = [freq_idx for freq_idx, freq in enumerate(freq_pts) if band[0] <= freq < band[1]]

        return band_ind, band_set, fft_ind

    
def _init_decoder_for_sim(n_steps = 10):
    kw = dict(control_method='fraction')
    sf = SmoothFilter(n_steps,**kw)
    ssm = train.endpt_2D_state_space
    units = [[23, 1],[24,1],[25,1]]
    decoder = One_Dim_LFP_Decoder(sf, units, ssm, binlen=0.1, n_subbins=1)
    learner = clda.DumbLearner()
        
    return decoder

def create_decoder(units, ssm, extractor_cls, extractor_kwargs, n_steps=10):
    kw = dict(control_method='fraction')
    sf = SmoothFilter(n_steps,**kw)
    decoder = One_Dim_LFP_Decoder(sf, units, ssm, extractor_cls, extractor_kwargs)

    return decoder








"""Simple Time-Sync

## Introduction
Eye tracking is often performed concurrently with external recording devices (e.g.
physiological sensors, motion capture) and/or software (e.g. for stimuli presentation,
experiment control). In order to correlate the eye tracking data with those of the
external devices/software, temporal alignment is of great importance.

By default, most devices/software come with their own clock to measure time
consistently. These clocks rarely share a common starting point which would allow for
automatic time alignment. However, they usually guarantee to be monotonicly increasing
and are very accurate when it comes to measuring time differences. Pupil Core software
is no exception in this regard.

In the example below, we show how one can calculate the offset between a custom "local"
clock and the clock used in Pupil Core software in real time.


## Local clocks
When syncing time, one needs to decide on which clock to sync to. This is often
dependent on the use case and the required measurement precision.

For this example, we define the "local" clock as `time.perf_counter()`, which is the
recommended clock for simple Python scripts [1]. However, you can sample a clock from
whatever device or software as long as it is monotonic and accurate when measuring time
differences. For example, experiment software like PsychoPy provide their own clock
functions [2].

If you need time measured in Unix epoch (less accurate but defines a common starting
point), use `time.time()` [3].

Important: Whatever clock function you use, make sure that you use it consistently
across the experiment. Any change to the clock, e.g. via a reset, invalidates the
measured clock offset and requires you to re-measure it. Since Pupil Core measures time
in seconds, this script expects the "local" clock to measure time in seconds as well.
[1] https://docs.python.org/3/library/time.html#time.perf_counter
[2] https://www.psychopy.org/api/clock.html
[3] https://docs.python.org/3/library/time.html#time.time


## Simplifying assumptions
For this script, we make a few simplifying assumptions. Specifically, we assume that
both involved clocks (local and pupil) do not drift. The term `drift` refers to the
issue that the clock offsets might change over time due to inaccuracies of the clocks.
A possible solution is to periodically measure the offset and correct for potential
drift over the course of a given recording. In this example, we measure the offset
once* and assume no drift on the offset. For a more accurate and stable time sync see
https://docs.pupil-labs.com/core/software/pupil-capture/#network-plugins
which is based on https://en.wikipedia.org/wiki/Network_Time_Protocol

*Below, we show an example of measuring the offset 10 times. This is actually to account
for variance in the network latency when measuring the offset, rather than clock drift.
To account for clock drift, one would have to implement periodic offset measurements
like in the Network_Time_Protocol link above.

## Requirements
- `time` module; built-in functionality; includes local clock function for this example
- `pyzmq` module; pip install pyzmq; handles network communication to Core software
"""
import time
import zmq


def main():
    """
    This example performs these steps:
    1. Setup network connection (always required)
    2. Setup "local" clock function (always required)
    3. Measure clock offset once
    4. Measure clock offset more reliably to account for network
    latency variance (multiple measurements)
    5. Infer remote pupil clock time from local clock measurement
    """

    # 1. Setup network connection (always required)
    socket = setup_pupil_remote_connection(ip_adress="127.0.0.1")

    # 2. Setup local clock function (always required)
    local_clock = time.time  # Unix time, less accurate
    print('locak_time',local_clock)
    # local_clock = psychopy.clock.MonotonicClock().getTime  # new psychopy clock
    # local_clock = existing_psychopy_clock.getTime  # existing psychopy clock
    # local_clock = time.perf_counter

    # 3. Measure clock offset once
    offset = measure_clock_offset(socket, clock_function=local_clock)
    print(f"Clock offset (1 measurement): {offset} seconds")

    # 4. Measure clock offset more reliably to account for network
    # latency variance (multiple measurements)
    number_of_measurements = 10
    stable_offset_mean = measure_clock_offset_stable(
        socket, clock_function=local_clock, nsamples=number_of_measurements
    )
    print(
        f"Mean clock offset ({number_of_measurements} measurements): "
        f"{stable_offset_mean} seconds"
    )

    # 5. Infer pupil clock time from "local" clock measurement
    local_time = local_clock()
    pupil_time = request_pupil_time(socket)
    pupil_time_calculated_locally = pupil_time - stable_offset_mean
    print(f"Local time: {local_time}")
    print(f"Pupil time (calculated locally): {pupil_time_calculated_locally}")


def setup_pupil_remote_connection(
    ip_adress: str = "127.0.0.1", port: int = 50020
) -> zmq.Socket:
    """Creates a zmq-REQ socket and connects it to Pupil Capture or Service
    See https://docs.pupil-labs.com/developer/core/network-api/ for details.
    """
    ctx = zmq.Context.instance()
    socket = ctx.socket(zmq.REQ)
    socket.connect(f"tcp://{ip_adress}:{port}")
    return socket


def request_pupil_time(socket):
    """Uses an existing Pupil Core software connection to request the remote time.
    Returns the current "pupil time" at the timepoint of reception.
    See https://docs.pupil-labs.com/core/terminology/#pupil-time for more information
    about "pupil time".
    """
    socket.send_string("t")
    pupil_time = socket.recv()
    return float(pupil_time)


def measure_clock_offset(socket, clock_function):
    """Calculates the offset between the Pupil Core software clock and a local clock.
    Requesting the remote pupil time takes time. This delay needs to be considered
    when calculating the clock offset. We measure the local time before (A) and
    after (B) the request and assume that the remote pupil time was measured at (A+B)/2,
    i.e. the midpoint between A and B.

    As a result, we have two measurements from two different clocks that were taken
    assumingly at the same point in time. The difference between them ("clock offset")
    allows us, given a new local clock measurement, to infer the corresponding time on
    the remote clock.
    """
    local_time_before = clock_function()
    pupil_time = request_pupil_time(socket)
    # print('pupil time = ',pupil_time)
    local_time_after = clock_function()

    local_time = (local_time_before + local_time_after) / 2.0
    clock_offset = pupil_time - local_time
    return clock_offset


def measure_clock_offset_stable(socket, clock_function, nsamples=10):
    """Returns the mean clock offset after multiple measurements to reduce the effect
    of varying network delay.

    Since the network connection to Pupil Capture/Service is not necessarily stable,
    one has to assume that the delays to send and receive commands are not symmetrical
    and might vary. To reduce the possible clock-offset estimation error, this function
    repeats the measurement multiple times and returns the mean clock offset.

    The variance of these measurements is expected to be higher for remote connections
    (two different computers) than for local connections (script and Core software
    running on the same computer). You can easily extend this function to perform
    further statistical analysis on your clock-offset measurements to examine the
    accuracy of the time sync.
    """
    assert nsamples > 0, "Requires at least one sample"
    offsets = [measure_clock_offset(socket, clock_function) for x in range(nsamples)]
    return sum(offsets) / len(offsets)  # mean offset


if __name__ == "__main__":
    main()


from django.contrib.staticfiles.storage import staticfiles_storage
from django.urls import reverse

from jinja2 import Environment

def environment(**options):
    env = Environment(**options)
    env.globals.update({
        'static': staticfiles_storage.url,
        'url': reverse,
    })
    return env



import re
import os
import json
import glob
from db.tracker import models

plexon = models.System.objects.get(name='plexon')
plex2 = models.System.objects.get(name='plexon2')
hdf = models.System.objects.get(name='hdf')

def reassoc_20130106():
    rplex = models.Feature.objects.get(name='relay_plexon')
    rplexb = models.Feature.objects.get(name='relay_plexbyte')

    for te in models.TaskEntry.objects.filter(feats__in=[rplex, rplexb]):
        plxdata = models.DataFile.objects.filter(system=plexon, entry=te)
        if len(plxdata) == 0:
            fname = te.plexfile()
            hdfdata = models.DataFile.objects.filter(system=hdf, entry=te)
            if fname is not None and len(hdfdata) == 1:
               path, ext = os.path.splitext(hdfdata[0].path)
               newname = os.path.join(plexon.path, path)+".plx"
               #os.rename(fname, newname)
               #models.DataFile(system=plexon, entry=te, path=path+'.plx').save()
               print(models.DataFile(system=plexon, entry=te, path=newname))

def move_fs_orphans(path='/storage/plexon/', dest='/storage/plexon/orphans'):
    valid = set(f.get_path() for f in models.DataFile.objects.filter(system=plexon))
    for f in glob.glob(path+"*.plx"):
        if f not in valid:
            fname = os.path.split(f)[1]
            os.rename(f, os.path.join(dest, fname))

def move_archive_orphans(path='/backup/bmi3d/plexon/', dest='/backup/bmi3d/plexon/orphans'):
    query = models.DataFile.objects.filter(system=plexon)
    files = set(os.path.join(path, df.path) for df in query)
    for plx in glob.glob(path+"*.plx"):
        if plx not in files:
            newname = os.path.join(dest, os.path.split(plx)[1])
            os.rename(plx, newname)
            print(newname)

def assoc_plexon2():
    rplex = models.Feature.objects.get(name='relay_plexon')
    rplexb = models.Feature.objects.get(name='relay_plexbyte')

    for te in models.TaskEntry.objects.filter(feats__in=[rplex, rplexb]):
        fname = te.plexfile("/storage/plexon2", True)
        if fname is not None:
            query = models.DataFile.objects.filter(system=hdf, entry=te)
            if len(query) == 0:
                raise Exception('invalid entry, unknown!')

            newname = os.path.splitext(query[0].path)[0]+'.plx'
            os.rename(fname, os.path.join("/storage/plexon2", newname))
            models.DataFile(system=plex2, entry=te, path=newname).save()
            print(newname)

def move_p2_orphans(path='/storage/plexon2/', dest='/storage/plexon2/orphans'):
    valid = set(f.get_path() for f in models.DataFile.objects.filter(system=plex2))
    for f in glob.glob(path+"*.plx"):
        if f not in valid:
            fname = os.path.split(f)[1]
            os.rename(f, os.path.join(dest, fname))

def set_archived():
    for df in models.DataFile.objects.all():
        if not os.path.isfile(df.get_path()):
            df.archived = True
            try:
                df.get_path()
                df.save()
            except IOError:
                print(df.path, "File not found!")

def old():
    path = "/storage/plexon/"
    plxform = re.compile('cart(\d{2})(\d{2})(\d{4})(\d{3}).plx')

    orphans = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f)) and plxform.match(f)]

    system = models.System.objects.get(name="plexon")
    for orphan in orphans:
        mtime = os.stat(os.path.join(path, orphan)).st_mtime
        day, month, year, num = plxform.match(orphan).groups()
        entries = models.TaskEntry.objects.filter(date__year=year, date__month=month, date__day=day)
        etime = dict([(e, i) for i, e in enumerate(sorted(entries, key=lambda x:x.date))])
        tdiffs = [(e, abs(mtime - json.loads(e.report)[-1][2])) if e.report != '[]' else e for e in entries]
        target, tdiff = sorted(tdiffs, key=lambda x:x[1])[0]
        if tdiff < 60:
            newname = "cart{time}_{num:02}.plx".format(time=target.date.strftime('%Y%m%d'), num=etime[target]+1)
            print("renaming %s to %s"%(os.path.join(path, orphan), os.path.join(path, newname)))
            df = models.DataFile(local=False, path=newname, system=system, entry=target)
            os.rename(os.path.join(path, orphan), os.path.join(path, newname))
            df.save()
            print(df)



'''
Tasks which control a plant under pure machine control. Used typically for initializing BMI decoder parameters.
'''
import numpy as np
import os
import tables
import time
import subprocess
import signal

from riglib.experiment import traits, Experiment
from riglib.bmi.state_space_models import StateSpaceEndptVel2D
from riglib.bmi.bmi import Decoder, BMILoop, MachineOnlyFilter
from riglib.bmi.extractor import DummyExtractor
from riglib.stereo_opengl.window import Window, WindowDispl2D

from built_in_tasks.manualcontrolmultitasks import ScreenTargetCapture
from built_in_tasks.bmimultitasks import BMIControlMulti

from .target_graphics import *

bmi_ssm_options = ['Endpt2D', 'Tentacle', 'Joint2L']

class EndPostureFeedbackController(BMILoop, traits.HasTraits):
    ssm_type_options = bmi_ssm_options
    ssm_type = traits.OptionsList(*bmi_ssm_options, bmi3d_input_options=bmi_ssm_options)
    decoder_update_rate = traits.Float(60, desc="Assist feedback rate (Hz)")

    def load_decoder(self):
        self.ssm = StateSpaceEndptVel2D()
        A, B, W = self.ssm.get_ssm_matrices()
        filt = MachineOnlyFilter(A, W)
        units = []
        self.decoder = Decoder(filt, units, self.ssm, binlen=1./self.decoder_update_rate)
        self.decoder.n_features = 1

    def create_feature_extractor(self):
        self.extractor = DummyExtractor()
        self._add_feature_extractor_dtype()


class TargetCaptureVisualFeedback(EndPostureFeedbackController, BMIControlMulti):
    assist_level = (1, 1)
    is_bmi_seed = True

    def move_effector(self):
        pass

class TargetCaptureVFB2DWindow(TargetCaptureVisualFeedback, WindowDispl2D):
    fps = 20.
    def __init__(self,*args, **kwargs):
        super(TargetCaptureVFB2DWindow, self).__init__(*args, **kwargs)
        self.assist_level = (1, 1)

    def _start_wait(self):
        self.wait_time = 0.
        super(TargetCaptureVFB2DWindow, self)._start_wait()

    def _test_start_trial(self, ts):
        return ts > self.wait_time and not self.pause

    @classmethod
    def get_desc(cls, params, report):
        if isinstance(report, list) and len(report) > 0:
            duration = report[-1][-1] - report[0][-1]
            reward_count = 0
            for item in report:
                if item[0] == "reward":
                    reward_count += 1
            return "{} rewarded trials in {} min".format(reward_count, int(np.ceil(duration / 60)))
        elif isinstance(report, dict):
            duration = report['runtime'] / 60
            reward_count = report['n_success_trials']
            return "{} rewarded trials in {} min".format(reward_count, int(np.ceil(duration / 60)))
        else:
            return "No trials"

from .target_graphics import target_colors

class TargetCaptureReplay(ScreenTargetCapture):
    '''
    Reads the frame-by-frame cursor and trial-by-trial target positions from a saved
    HDF file to display an exact copy of a previous experiment. 
    Doesn't really work, do not recommend using this.
    '''

    hdf_filepath = traits.String("", desc="Filepath of hdf file to replay")

    exclude_parent_traits = list(set(ScreenTargetCapture.class_traits().keys()) - \
        set(['window_size', 'fullscreen']))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.t0 = time.perf_counter()
        with tables.open_file(self.hdf_filepath, 'r') as f:
            task = f.root.task.read()
            state = f.root.task_msgs.read()
            trial = f.root.trials.read()
            params = f.root.task.attrs._f_list("user")
            self.task_meta = {k : getattr(f.root.task.attrs, k) for k in params}
        self.replay_state = state
        self.replay_task = task
        self.replay_trial = trial
        for k, v in self.task_meta.items():
            if k in self.exclude_parent_traits:
                print("setting {} to {}".format(k, v))
                setattr(self, k, v)

        # Have to additionally reset the targets since they are created in super().__init__()
        target1 = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])
        target2 = VirtualCircularTarget(target_radius=self.target_radius, target_color=target_colors[self.target_color])
        self.targets = [target1, target2]

    def _test_start_trial(self, time_in_state):
        '''Wait for the state change in the HDF file in case there is autostart enabled'''
        trials = self.replay_state[self.replay_state['msg'] == b'target']
        upcoming_trials = [t['time']-1 for t in trials if self.replay_task[t['time']]['trial'] >= self.calc_trial_num()]
        return (np.array(upcoming_trials) <= self.cycle_count).any()

    def _parse_next_trial(self):
        '''Ignore the generator'''
        self.targs = []
        self.gen_indices = []
        trial_num = self.calc_trial_num()
        for trial in self.replay_trial:
            if trial['trial'] == trial_num:
                self.targs.append(trial['target'])
                self.gen_indices.append(trial['index'])

    def _cycle(self):
        '''Have to fudge the cycle_count a bit in case the fps isn't exactly the same'''
        super()._cycle()
        t1 = time.perf_counter() - self.t0
        self.cycle_count = int(t1*self.fps)

    def move_effector(self):
        current_pt = self.replay_task['cursor'][self.cycle_count]
        self.plant.set_endpoint_pos(current_pt)

    def _test_stop(self, ts):
        return super()._test_stop(ts) or self.cycle_count == len(self.replay_task)


class YouTube(Experiment):

    youtube_url = traits.String("", desc="URL pointing to a YouTube video. Only works for videos that support embedding")

    def start_video(self):
        self.video_process = subprocess.Popen(["bash", "../utils/start-youtube.sh", self.youtube_url])

    def stop_video(self):
        os.kill(self.video_process.pid, signal.SIGINT)
        self.video_process.wait()
        
    def _cycle(self):
        try:
            status = self.video_process.poll()
            if status is not None:
                self.state = None
        except:
            pass
        super()._cycle()
   
    def run(self):
        '''
        Code to execute immediately prior to the beginning of the task FSM executing, or after the FSM has finished running. 
        See riglib.experiment.Experiment.run(). Starts the YouTube video and stops it after the FSM has finished running
        '''
        try:
            self.start_video()
            super().run()
        finally:
            print("Stopping video")
            self.stop_video()
            



'''
Screen Space Ambient Occlusion (SSAO) Renderer

This module implements a Screen Space Ambient Occlusion renderer, which is a technique
used to approximate ambient occlusion in real-time. SSAO adds depth and realism to 3D
scenes by darkening areas that are occluded or surrounded by other geometry.

The SSAO class extends FBOrender to provide a multi-pass rendering pipeline:
1. Render scene to obtain normal and depth information
2. Calculate ambient occlusion based on the normal and depth data
3. Apply the ambient occlusion to the final render
'''

import numpy as np
from OpenGL.GL import *

from .render import Renderer
from .fbo import FBOrender, FBO
from ..textures import Texture

class SSAO(FBOrender):
    def __init__(self, *args, sf=1, **kwargs):
        super(SSAO, self).__init__(*args, **kwargs)
        self.sf = sf
        w, h = self.size[0] / self.sf, self.size[1] / self.sf
        
        self.normdepth = FBO(["color0", "depth"], size=(w,h))
        self.ping = FBO(['color0'], size=(w,h))
        self.pong = FBO(["color0"], size=(w,h))

        self.add_shader("fsquad", GL_VERTEX_SHADER, "fsquad.v.glsl")
        self.add_shader("ssao_pass1", GL_FRAGMENT_SHADER, "ssao_pass1.f.glsl")
        self.add_shader("ssao_pass2", GL_FRAGMENT_SHADER, "ssao_pass2.f.glsl")
        self.add_shader("ssao_pass3", GL_FRAGMENT_SHADER, "ssao_pass3.f.glsl", "phong.f.glsl")
        self.add_shader("hblur", GL_FRAGMENT_SHADER, "hblur.f.glsl")
        self.add_shader("vblur", GL_FRAGMENT_SHADER, "vblur.f.glsl")

        #override the default shader with this passthru + ssao_pass1 to store depth
        self.add_program("ssao_pass1", ("passthru", "ssao_pass1"))
        self.add_program("ssao_pass2", ("fsquad", "ssao_pass2"))
        self.add_program("hblur", ("fsquad", "hblur"))
        self.add_program("vblur", ("fsquad", "vblur"))
        self.add_program("ssao_pass3", ("passthru", "ssao_pass3"))

        # Some debug shaders
        # self.add_shader("fsquad", GL_VERTEX_SHADER, "fsquad.v.glsl")
        # self.add_shader("fsquad_frag", GL_FRAGMENT_SHADER, "fsquad.f.glsl")
        # self.add_program("none", ("fsquad", "fsquad_frag"))

        randtex = np.random.rand(3, int(w), int(h))
        randtex /= randtex.sum(0)
        self.rnm = Texture(randtex.T, wrap_x=GL_REPEAT, wrap_y=GL_REPEAT, 
            magfilter=GL_NEAREST, minfilter=GL_NEAREST)
        self.rnm.init()

        self.clips = args[2], args[3]

    def draw(self, root, **kwargs):
        # Save the current viewport
        original_viewport = glGetIntegerv(GL_VIEWPORT)
        original_framebuffer = glGetIntegerv(GL_FRAMEBUFFER_BINDING)

        # Set the new viewport for SSAO calculations
        new_viewport = (0, 0, self.size[0]//self.sf, self.size[1]//self.sf)
        glViewport(*new_viewport)

        # First, draw the whole scene, but only read the normals and depth into ssao
        self.draw_to_fbo(self.normdepth, root, shader="ssao_pass1", apply_default=True, **kwargs)
        
        # Now, do the actual ssao calculations, and draw it into pong
        self.draw_fsquad_to_fbo(self.pong, "ssao_pass2", rnm=self.rnm,
            normalMap=self.normdepth['color0'], depthMap=self.normdepth['depth'],
            nearclip=self.clips[0], farclip=self.clips[1] )
        
        # Blur the textures
        self.draw_fsquad_to_fbo(self.ping, "hblur", tex=self.pong['color0'], blur=1./(self.size[0]/self.sf))
        self.draw_fsquad_to_fbo(self.pong, "vblur", tex=self.ping['color0'], blur=1./(self.size[0]/self.sf))
        
        # glViewport(*original_viewport)
        # glBindFramebuffer(GL_FRAMEBUFFER, original_framebuffer)
        # self.draw_fsquad("none", tex=self.pong['color0'])
        # return

        # Restore the original viewport
        glViewport(*original_viewport)
        glBindFramebuffer(GL_FRAMEBUFFER, original_framebuffer)
        super(SSAO, self).draw(root, shader="ssao_pass3", apply_default=True, shadow=self.pong['color0'], 
            window=[float(i) for i in original_viewport], **kwargs)
        super(SSAO, self).draw(root, shader="ui", **kwargs)
        
    def clear(self):
        self.normdepth.clear()
        self.ping.clear()
        self.pong.clear()
    
    def draw_done(self):
        super(SSAO, self).draw_done()
        self.clear()



import time
from riglib import source
from riglib.ecube import Broadband, LFP, File, LFP_Blanking_File, LFP_Blanking, MultiSource, MultiSource_File, make_source_class
from riglib.ecube.file import parse_file
from riglib.bmi import state_space_models, train, extractor
import numpy as np
import unittest

import unittest

STREAMING_DURATION = 3

class TestStreaming(unittest.TestCase):

    # Note: for the multiprocessing code to work, you must have servernode-control open separately

    @unittest.skip('works')
    def test_ecube_stream(self):
        channels = [1, 3]
        bb = Broadband(channels=channels) # Note: this is not how this is normally used, just testing
        bb.start()
        data = bb.conn.get()
        bb.stop()
        print(data[2].shape)
    
    @unittest.skip('works')
    def test_broadband_class(self):
        channels = [1, 3]
        bb = Broadband(channels=channels)
        bb.start()
        ch_data = []
        for d in range(2):
            for i in range(len(channels)):
                ch, data = bb.get()
                ch_data.append(data)
                print(f"Got channel {ch} with {data.shape} samples")
        bb.stop()

    @unittest.skip('works')
    def test_broadband_datasource(self):
        channels = [1, 62]
        ds = source.MultiChanDataSource(Broadband, channels=channels)
        ds.start()
        time.sleep(STREAMING_DURATION)
        data = ds.get_new(channels)
        ds.stop()
        data = np.array(data)

        n_samples = int(Broadband.update_freq * STREAMING_DURATION / 728) * 728 # closest multiple of 728 (floor)

        self.assertEqual(data.shape[0], len(channels))
        self.assertEqual(data.shape[1], n_samples)
    
    @unittest.skip('works')
    def test_ds_with_extractor(self):
        # Create the datasource
        channels = [1, 62]
        ds = source.MultiChanDataSource(LFP, channels=channels)

        # Make a feature extractor
        extr = extractor.LFPMTMPowerExtractor(ds, channels=channels, bands=[(90,110)], win_len=0.1, fs=1000)

        # Run the feature extractor
        extract_rate = 1/60
        feats = []
        ds.start()
        t0 = time.perf_counter()
        while (time.perf_counter() - STREAMING_DURATION < t0):
            neural_features_dict = extr(time.perf_counter())
            feats.append(neural_features_dict['lfp_power'])
            time.sleep(extract_rate)
        
        ds.stop()
        data = np.array(feats)
        print(feats)
        print(data.shape)
        self.assertEqual(data.shape[1], len(channels))
        self.assertEqual(data.shape[0], STREAMING_DURATION/extract_rate)

    @unittest.skip('works')
    def test_lfp_blanking(self):
        channels = [1, 3]
        file = "/media/server/raw/ecube/2022-12-23_BMI3D_te7797"
        bb = LFP_Blanking_File(channels=channels, ecube_bmi_filename=file, trig_channel=0) # Note: this is not how this is normally used, just testing
        bb.start()
        ch_data = []
        chs = []
        for d in range(100):
            for i in range(len(channels)):
                ch, data = bb.get()
                ch_data.append(data)
                chs.append(ch)
                if np.count_nonzero(np.isnan(data)):
                    print(f"Got some nans in ch {ch}")
        print(np.unique(chs))
        bb.stop()

    @unittest.skip("works") 
    def test_multisource(self):
        channels = [1, 33, 34, 97]
        # file = "/Users/leoscholl/2023-10-10_BMI3D_te11503"
        bb = MultiSource_File(channels=channels)
        bb.start()
        ch_data = []
        chs = []
        for d in range(100):
            print('.', end="")
            for i in range(len(channels)):
                ch, data = bb.get()
                ch_data.append(data)
                chs.append(ch)
                if np.count_nonzero(np.isnan(data)):
                    print(f"Got some nans in ch {ch}")
        print(np.unique(chs))
        bb.stop()

    # @unittest.skip("doesn't work on my mac - just a bunch of zeros")
    def test_multisource_ds(self):
        channels = [1, 40+32, 97]
        # file = "/Users/leoscholl/2023-10-10_BMI3D_te11503"
        ds = source.MultiChanDataSource(MultiSource_File, channels=channels)
        ds.start()
        time.sleep(2*STREAMING_DURATION)
        data = ds.get_new(channels)
        print(data)
        ds.stop()

        n_samples = int(MultiSource_File.update_freq * STREAMING_DURATION / 728) * 728 # closest multiple of 728 (floor)

        self.assertEqual(len(data), len(channels))
        # self.assertEqual(len(data[0]), n_samples)
        # self.assertEqual(len(data[1]), len(data[2]))

        from matplotlib import pyplot as plt
        plt.plot(data[1])
        plt.show()

class TestFileLoadin(unittest.TestCase):

    def test_ecube_load(self):
        pass


    def extract_from_file_lfp(self):

        pass

        #LFPMTMPowerExtractor.extract_from_file()

if __name__ == '__main__':
    unittest.main()


'''
Classes for BMI decoding using the Steady-state Kalman filter. Based
heavily on the kfdecoder module. 
'''
import numpy as np
from . import bmi
from . import kfdecoder

class SteadyStateKalmanFilter(bmi.GaussianStateHMM):
    """
    Low-level KF in steady-state

    Model: 
       x_{t+1} = Ax_t + w_t;   w_t ~ N(0, W)
           y_t = Cx_t + q_t;   q_t ~ N(0, Q)
    """ 
    model_attrs = ['F', 'K']

    def __init__(self, *args, **kwargs):
        '''Docstring

        Parameters
        ----------

        Returns
        -------

        '''
        if len(list(kwargs.keys())) == 0:
            ## This condition should only be true in the unpickling phase
            pass
        else:
            if 'A' in kwargs and 'C' in kwargs and 'W' in kwargs and 'Q' in kwargs:
                A = kwargs.pop('A')
                C = kwargs.pop('C')
                W = kwargs.pop('W')
                Q = kwargs.pop('Q')
                kf = kfdecoder.KalmanFilter(A, W, C, Q, **kwargs)
                F, K = kf.get_sskf()
            elif 'F' in kwargs and 'K' in kwargs:
                F = kwargs['F']
                K = kwargs['K']
            self.F = F
            self.K = K
            self._pickle_init()


    def _init_state(self, init_state=None, init_cov=None):                     
        """                                                                    
        Initialize the state of the filter with a mean and covariance (uncertainty)
        Docstring

        Parameters
        ----------

        Returns
        -------
        """                                                                    
        ## Initialize the BMI state, assuming                                  
        nS = self.n_states                                                 
        if init_state == None:                                                 
            init_state = np.mat( np.zeros([nS, 1]) )                           
            if self.include_offset: init_state[-1,0] = 1                       
        if init_cov == None:                                                   
            init_cov = np.mat( np.zeros([nS, nS]) )
        self.state = bmi.GaussianState(init_state, init_cov)                       

    def _pickle_init(self):
        '''Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        nS = self.F.shape[0]
        self.I = np.mat(np.eye(nS))

    def get_sskf(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        return self.F, self.K

    def _forward_infer(self, st, obs_t, Bu=None, u=None, target_state=None, 
                       obs_is_control_independent=True, bias_comp=False, **kwargs):
        '''
        Estimate p(x_t | ..., y_{t-1}, y_t)
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        F, K = self.F, self.K
        if Bu is not None:
            post_state_mean = F*st.mean + K*obs_t + Bu
        else:
            post_state_mean = F*st.mean + K*obs_t

        I = self.I
        post_state = I*st # Force memory reallocation for the Gaussian
        post_state.mean = post_state_mean
        return post_state

    def __getstate__(self):
        '''
        Pickle only the F and the K matrices
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        return dict(F=self.F, K=self.K)
    
    @property
    def n_states(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        return self.F.shape[0]

    @property
    def include_offset(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        return np.all(self.F[-1, :-1] == 0) and (self.F[-1, -1] == 1)

    def get_K_null(self):
        '''
        $$y_{null} = K_{null} * y_t$$ gives the "null" component of the spike inputs, i.e. $$K_t*y_{null} = 0_{N\times 1}$$
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        K = np.mat(self.K)
        n_neurons = K.shape[1]
        K_null = np.eye(n_neurons) - np.linalg.pinv(K) * K
        return K_null

class SSKFDecoder(bmi.Decoder, bmi.BMI):
    ''' Docstring '''
    pass



'''
Interface between the Django database methods/models and data analysis code
'''
# django initialization
import os
os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"
from .boot_django import boot_django
from django.db import connection
from django.db.utils import OperationalError

import sys
import json
import numpy as np
import datetime
import pickle
import tables
from collections import defaultdict

from .tracker import models

# configure django
boot_django()

# decorator for mysql access
def mysqlreconnect(f):
    def inner(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except OperationalError:
            connection.close()
            return f(*args, **kwargs)
    return inner

class TaskEntry(object):
    '''
    Wrapper class for the TaskEntry django class so that object-oriented methods
    can be defined for TaskEntry blocks (e.g., for analysis methods for a particular experiment)
    without needing to modfiy the database model.
    '''
    def __init__(self, task_entry_id, dbname='default', **kwargs):
        '''
        Constructor for TaskEntry

        Parameters
        ----------
        task_entry_id : int
            'id' number associated with the TaskEntry (database primary key)
        dbname : string, optional, default='default'
            Name of database (if multiple databases are available). Database must
            be declared in db/settings.py
        **kwargs : dict
            For any additional parameters unused by child classes

        Returns
        -------
        TaskEntry instance
        '''
        self.dbname = dbname
        if isinstance(task_entry_id, models.TaskEntry):
            self.record = task_entry_id
        else:
            self.record = models.TaskEntry.objects.using(self.dbname).get(id=task_entry_id)
        self.id = self.record.id
        self.params = self.record.params
        if (isinstance(self.params, str) or isinstance(self.params, str)) and len(self.params) > 0:
            self.params = json.loads(self.record.params)

            # Add the params dict to the object's dict
            for key, value in list(self.params.items()):
                try:
                    setattr(self, key, value)
                except AttributeError: # if the object already has an attribute of the same name ...
                    setattr(self, key + '_param', value)
                except:
                    import traceback
                    traceback.print_exc()

                if isinstance(value, list):
                    self.params[key] = tuple(value)

        self.date = self.record.date

        ## Extract a date month-day-year date for determining if other blocks were on the same day
        self.calendar_date = datetime.datetime(self.date.year, self.date.month, self.date.day)

        self.notes = self.record.notes
        self.subject = models.Subject.objects.using(dbname).get(pk=self.record.subject_id).name

        # Load decoder record
        if 'decoder' in self.params:
            self.decoder_record = models.Decoder.objects.using(self.record._state.db).get(pk=self.params['decoder'])
        elif 'bmi' in self.params:
            self.decoder_record = models.Decoder.objects.using(self.record._state.db).get(pk=self.params['bmi'])
        else: # Try direct lookup
            try:
                self.decoder_record = models.Decoder.objects.using(self.record._state.db).get(entry_id=self.id)
            except:
                self.decoder_record = None

        # Load the event log (report)
        try:
            self.report = json.loads(self.record.report)
        except:
            self.report = ''


    def get_decoders_trained_in_block(self, return_type='record'):
        '''
        Retrieve decoders associated with this block. A block may have multiple decoders
        associated with it, e.g., a single decoder seeding block may be used to generate several seed decoders

        Parameters
        ----------
        return_type: string
            'record' means the Django database records are returned. 'object' means the un-pickled riglib.bmi.Decoder objects are returned

        Returns
        -------
        list or object
            If only one decoder is linked to this task entry, an object is returned (either a db record or a Decoder instance). If multiple decoders are linked,
            a list of objects is returned
        '''
        records = models.Decoder.objects.using(self.dbname).filter(entry_id=self.id)
        if return_type == 'record':
            decoder_objects = list(records)
        elif return_type == 'object':
            decoder_objects = [x.load() for x in records]
        else:
            raise ValueError("Unrecognized return_type!")
        if len(decoder_objects) == 1: decoder_objects = decoder_objects[0]
        return decoder_objects

    def summary_stats(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        report = self.record.offline_report()
        for key in report:
            print(key, report[key])

    def proc(self, filt=None, proc=None, cond=None, comb=None, **kwargs):
        '''
        Generic trial-level data analysis function

        Parameters
        ----------
        filt: callable; call signature: trial_filter_fn(trial_msgs)
            Function must return True/False values to determine if a set of trial messages constitutes a valid set for the analysis
        proc: callable; call signature: trial_proc_fn(task_entry, trial_msgs)
            The main workhorse function
        cond: callable; call signature: trial_condition_fn(task_entry, trial_msgs)
            Determine what the trial *subtype* is (useful for separating out various types of catch trials)
        comb: callable; call signature: data_comb_fn(list)
            Combine the list into the desired output structure
        kwargs: optional keyword arguments
            For 'legacy' compatibility, you can also specify 'trial_filter_fn' for 'filt', 'trial_proc_fn' for 'proc',
            'trial_condition_fn' for 'cond', and 'data_comb_fn' for comb. These are ignored if any newer equivalents are specified.
            All other keyword arguments are passed to the 'proc' function.

        Returns
        -------
        result: list
            The results of all the analysis. The length of the returned list equals len(self.blocks). Sub-blocks
            grouped by tuples are combined into a single result.

        '''
        if filt == None:
            filt = kwargs.pop('trial_filter_fn', default_trial_filter_fn)
        if cond == None:
            cond = kwargs.pop('trial_condition_fn', default_trial_condition_fn)
        if proc == None:
            proc = kwargs.pop('trial_proc_fn', default_trial_proc_fn)
        if comb == None:
            comb = kwargs.pop('data_comb_fn', default_data_comb_fn)


        trial_filter_fn = filt
        trial_condition_fn = cond
        trial_proc_fn = proc
        data_comb_fn = comb

        # import trial_filter_functions, trial_proc_functions, trial_condition_functions
        # if isinstance(trial_filter_fn, str):
        #     trial_filter_fn = getattr(trial_filter_functions, trial_filter_fn)

        # if isinstance(trial_proc_fn, str):
        #     trial_proc_fn = getattr(trial_proc_functions, trial_proc_fn)

        # if isinstance(trial_condition_fn, str):
        #     trial_condition_fn = getattr(trial_condition_functions, trial_condition_fn)

        te = self
        trial_msgs = [msgs for msgs in te.trial_msgs if trial_filter_fn(te, msgs)]
        n_trials = len(trial_msgs)

        blockset_data = defaultdict(list)

        ## Call a function on each trial
        for k in range(n_trials):
            output = trial_proc_fn(te, trial_msgs[k], **kwargs)
            trial_condition = trial_condition_fn(te, trial_msgs[k])
            blockset_data[trial_condition].append(output)

        newdata = dict()
        for key in blockset_data:
            newdata[key] = data_comb_fn(blockset_data[key])

        if len(list(newdata.keys())) == 1:
            key = list(newdata.keys())[0]
            return newdata[key]
        else:
            return newdata

    def get_cached_attr(self, key, fn, clean=False):
        '''
        Generic method for saving the results of a computation to the supplementary_data_file associated with this block

        Parameters
        ----------
        key : string
            Variable name in cache file
        fn : callable
            Function to execute (no arguments) to get the required data, if it is not present in the file
        clean : bool, default=False
            If true, force the recomputation of the data product even if it is already present in the cache file
        '''
        from scipy.io import savemat, loadmat
        if hasattr(self, '_%s' % key):
            return getattr(self, '_%s' % key)
        try:
            task_supplement = models.System.objects.get(name='task_supplement')
            supplementary_data_file = os.path.join(task_supplement.path, '%d.mat' % self.record.id)
        except:
            import warnings
            warnings.warn('No "task_supplement" system. Please add one.')
            supplementary_data_file = None
        if supplementary_data_file is None:
            data = fn()
            setattr(self, '_%s' % key, data)
            return getattr(self, '_%s' % key)
        if not os.path.exists(supplementary_data_file):
            data = fn()
            savemat(supplementary_data_file, dict(key=data))
            setattr(self, '_%s' % key, data)
            return getattr(self, '_%s' % key)
        else:
            supplementary_data = loadmat(supplementary_data_file)
            if (not key in supplementary_data) or clean:
                supplementary_data[key] = fn()
                savemat(supplementary_data_file, supplementary_data)
            setattr(self, '_%s' % key, supplementary_data[key])
            return getattr(self, '_%s' % key)

    def get_matching_state_transition_seq(self, seq):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        task_msgs = self.hdf.root.task_msgs[:]
        seq = np.array(seq, dtype='|S256')
        msg_list_inds = []
        trial_msgs = []
        epochs = []
        for k in range(len(task_msgs)-len(seq)):
            if np.all(task_msgs[k:k+len(seq)]['msg'] == seq):
                msg_list_inds.append(k)
                trial_msgs.append(task_msgs[k:k+len(seq)])
                epochs.append((task_msgs[k]['time'], task_msgs[k+len(seq)-1]['time']))
        epochs = np.vstack(epochs)
        return msg_list_inds, trial_msgs, epochs

    def get_task_var_during_epochs(self, epochs, var_name, comb_fn=lambda x:x, start_offset=0):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        data = []
        for st, end in epochs:
            st += start_offset
            epoch_data = self.hdf.root.task[st:end][var_name]

            # Determine whether all the rows in the extracted sub-table are the same
            # This code is stolen from the interweb:
            # http://stackoverflow.com/questions/16970982/find-unique-rows-in-numpy-array
            a = epoch_data
            b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
            unique_epoch_data = np.unique(b).view(a.dtype).reshape(-1, a.shape[1])

            # unique_epoch_data = np.unique(epoch_data.ravel())
            if len(unique_epoch_data) == 1:
                data.append(unique_epoch_data)
            else:
                data.append(epoch_data)
        return comb_fn(data)

    @property
    def hdf_filename(self):
        '''
        Get the task-generated HDF file linked to this TaskEntry
        '''
        q = models.DataFile.objects.using(self.record._state.db).filter(entry_id=self.id, system__name='hdf')
        if len(q) == 0:
            # empty string for HDF filename if none is linked in the database
            return ''
        elif len(q) == 1:
            q = q[0]
            # dbconfig = getattr(config, 'db_config_%s' % self.record._state.db)
            # data_path = models.KeyValueStore.get('data_path', default='/storage', dbname=self.record._state.db)
            return os.path.join(q.system.path, q.path)

    @property
    def hdf(self):
        '''
        Return a reference to the HDF file recorded during this TaskEntry
        '''
        if not hasattr(self, 'hdf_file'):
            self.hdf_file = tables.open_file(self.hdf_filename, mode='r')
        return self.hdf_file

    def close_hdf(self):
        if hasattr(self, 'hdf_file'):
            self.hdf_file.close()
            delattr(self, 'hdf_file')

    def close(self):
        '''
        Cleanup functions
        '''
        self.close_hdf()

    @property
    def plx(self):
        '''
        Return a reference to the opened plx file recorded during this TaskEntry
        '''
        try:
            self._plx
        except:
            from plexon import plexfile
            self._plx = plexfile.openFile(str(self.plx_filename))

            # parse out events
            from riglib.dio import parse
            self.strobe_data = parse.parse_data(self._plx.events[:].data)
        return self._plx

    @property
    def task(self):
        '''
        Return database record of the task performed during this TaskEntry
        '''
        return self.record.task

    @property
    def decoder(self):
        '''
        Return a reference to the unpickled riglib.bmi.Decoder instance associated with this TaskEntry.
        This function will error if there is no decoder actually associated with this TaskEntry
        '''
        if not hasattr(self, '_decoder_obj'):
            self._decoder_obj = self.decoder_record.load(encoding='latin1')
        return self._decoder_obj

    @property
    def clda_param_hist(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        bmi_params = models.System.objects.get(name='bmi_params')
        q = models.DataFile.objects.filter(entry_id=self.id, system_id=bmi_params.id)

        if not hasattr(self, '_clda_param_hist'):
            if hasattr(self.hdf.root, 'clda'):
                self._clda_param_hist = self.hdf.root.clda[:]
            elif len(q) > 0: #get_bmiparams_file(self.record) is not None:
                fname = os.path.join(q[0].system.path, q[0].path)
                self._clda_param_hist = np.load(fname)
            else:
                self._clda_param_hist = None
        return self._clda_param_hist

    @property
    def length(self):
        '''
        Return the length of the TaskEntry, in seconds. Only works for tasks which inherit from LogExperiment, as it uses the event log for determining timing.
        '''
        try:
            report = json.loads(self.record.report)
        except:
            return 0.0
        return report[-1][2]-report[0][2]

    def get_datafile(self, system_name, **query_kwargs):
        '''
        Look up the file linked to this TaskEntry from a specific system

        Parameters
        ----------
        system_name : string
            Name of system (must match the database record name)
        intermediate_path: string
            specific directory structure for this system. Some systems have inconsistent directory structures
        **query_kwargs: keyword arguments
            These are passed to the Django record 'filter' function

        Returns
        -------
        filename of associated file
        '''
        file_records = models.DataFile.objects.using(self.record._state.db).filter(entry_id=self.id, system__name=system_name, **query_kwargs)
        dbconfig = getattr(config, 'db_config_%s' % self.record._state.db)
        file_names = [os.path.join(q.system.path, q.path) for q in file_records]
        if len(file_names) == 1:
            return file_names[0]
        else:
            return list(file_names)

    @property
    def plx_filename(self):
        '''
        Return the name of the plx file associated with this TaskEntry
        '''
        return self.get_datafile('plexon')

    @property
    def plx2_filename(self):
        '''
        Returns the name of any files associated with the plexon2 system. Used briefly only in a rig where there were multiple recording systems used simultaneously.
        '''
        return self.get_datafile('plexon2')

    @property
    def bmiparams_filename(self):
        '''
        Return the name of the npz file that was used to store CLDA parameters, if one exists
        '''
        return self.get_datafile('bmi_params')

    @property
    def blackrock_filenames(self):
        x = self.get_datafile(system_name='blackrock')
        y = self.get_datafile(system_name='blackrock2')

        if type(x) is not list:
            x = [x]
        if type(y) is not list:
            y = [y]

        return x+y
    @property
    def nev_filename(self):
        '''
        Get any blackrock nev files associated with this TaskEntry
        '''
        return self.get_datafile(system_name='blackrock', path__endswith='.nev')

    @property
    def nsx_filenames(self):
        '''
        Get any blackrock nsx files associated with this TaskEntry
        '''
        return self.get_datafile(system_name='blackrock', path__endswith='.nsx')

    @property
    def decoder_filename(self):
        '''
        Get filename of riglib.bmi.Decoder object associate with this TaskEntry. Only works for BMI tasks!
        '''
        return self.decoder_record.filename #get_datafile('bmi', intermediate_path='rawdata')
        # decoder_basename = self.decoder_record.path
        # dbconfig = getattr(config, 'db_config_%s' % self.record._state.db)
        # return os.path.join(dbconfig['data_path'], 'decoders', decoder_basename)

    @property
    def name(self):
        '''
        Return the 'name' of the TaskEntry used to set all the names of linked data files
        '''
        # TODO this needs to be hacked because the current way of determining a
        # a filename depends on the number of things in the database, i.e. if
        # after the fact a record is removed, the number might change. read from
        # the file instead
        return str(os.path.basename(self.hdf_filename).rstrip('.hdf'))

    def trained_decoder_filenames(self):
        decoder_records = self.get_decoders_trained_in_block()
        filenames = []
        if np.iterable(decoder_records):
            for rec in decoder_records:
                filenames.append(rec.filename)
        else:
            filenames.append(decoder_records.filename)
        return filenames

    def __str__(self):
        return self.record.__str__()

    def __repr__(self):
        return self.record.__repr__()

    @property
    def task_type(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        if 'bmi' in self.task.name:
            return 'BMI'
        elif 'clda' in self.task.name:
            return 'CLDA'

    @property
    def n_rewards(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        # of rewards given during a block. This number could be different
        from the total number of trials if children of this class over-ride
        the n_trials calculator to exclude trials of a certain type, e.g. BMI
        trials in which the subject was assiste
        '''
        return self.record.offline_report()['Total rewards']

    @property
    def total_reward_time(self):
        '''
        Docstring

        Parameters
        ----------

        Returns
        -------
        '''
        return self.n_rewards * self.reward_time

    @property
    def datafiles(self):
        '''
        Returns a dictionary where keys are system names and values are the location(s) of associated data files
        '''
        datafiles = models.DataFile.objects.using(self.record._state.db).filter(entry_id=self.id)
        files = dict((d.system.name, d.get_path()) for d in datafiles)
        return files

    def save_to_database(self, new_db):
        '''
        Save the current task entry (and associated records) to a different database
        '''
        current_db = self.record._state.db
        if current_db == new_db:
            print("new database and current database are the same!")
            return

        # save the subject
        self.record.subject.save(using=new_db)

        # save the task
        self.record.task.save(using=new_db)

        # save the task entry record
        self.record.save(using=new_db)

        # save the records of the datafiles associated with the task entry
        # A bit of circuswork is required to move the DataFiles because they have a foreign key for the system that correspomds to the datafile
        datafiles = models.DataFile.objects.using(current_db).filter(entry_id=self.id)
        for d in datafiles:
            sys_name = d.system.name
            d.save(using=new_db)
            new_sys = models.System.objects.using(new_db).get(name=d.system.name)
            d.system = new_sys
            d.save()

        # save any decoder records used by this block
        if not (self.decoder_record is None):
            self.decoder_record.save(using=new_db)

    def get_state_inds(self, state_name):
        '''
        Determine the rows of the task table in which the task is in a particular state.
        '''
        task_msgs = self.hdf.root.task_msgs[:]
        n_msgs = len(task_msgs)
        msg_inds = np.array([k for k in range(n_msgs) if task_msgs[k]['msg'] == state_name])

        inds = np.zeros(len(self.hdf.root.task), dtype=bool)
        for idx in msg_inds:
            if task_msgs[idx+1]['msg'] == 'None':
                break
            for k in range(task_msgs[idx]['time'], task_msgs[idx+1]['time']+1):
                inds[k] = 1
        return inds



class TaskEntryCollection(object):
    '''
    Container for analyzing multiple task entries with an arbitrarily deep hierarchical structure
    '''
    def __init__(self, blocks, name='', cls=TaskEntry, **kwargs):
        '''
        Constructor for TaskEntryCollection

        Parameters
        ----------
        blocks : np.iterable
            Some iterable object which contains TaskEntry ID numbers to look up in the database
        name : string, optional, default=''
            Name to give this collection
        cls : type, optional, default=TaskEntry
            class constructor to use for all the IDs. Uses the generic TaskEntry defined above by default

        Returns
        -------
        '''
        self.block_ids = blocks
        self.blocks = parse_blocks(blocks, cls=cls, **kwargs)
        self.kwargs = kwargs
        self.name = name

    def __len__(self):
        return len(self.blocks)

    def proc_trials(self, filt=None, proc=None, cond=None, comb=None, verbose=False, max_errors=10, **kwargs):
        '''
        Generic framework to perform a trial-level analysis on the entire dataset

        Parameters
        ----------
        filt: callable; call signature: trial_filter_fn(trial_msgs)
            Function must return True/False values to determine if a set of trial messages constitutes a valid set for the analysis
        proc: callable; call signature: trial_proc_fn(task_entry, trial_msgs)
            The main workhorse function
        cond: callable; call signature: trial_condition_fn(task_entry, trial_msgs)
            Determine what the trial *subtype* is (useful for separating out various types of catch trials)
        comb: callable; call signature: data_comb_fn(list)
            Combine the list into the desired output structure
        verbose: boolean, optional, default = True
            Feedback print statements so that you know processing is happening
        max_errors: int, optional, default = 10
            Number of trials resulting in error before the processing quits. Below this threshold, errors are printed but the code continues on to the next trial.

        Returns
        -------
        result: list
            The results of all the analysis. The length of the returned list equals len(self.blocks). Sub-blocks
            grouped by tuples are combined into a single result.
        '''

        if filt == None:
            filt = kwargs.pop('trial_filter_fn', default_trial_filter_fn)
        if cond == None:
            cond = kwargs.pop('trial_condition_fn', default_trial_condition_fn)
        if proc == None:
            proc = kwargs.pop('trial_proc_fn', default_trial_proc_fn)
        if comb == None:
            comb = kwargs.pop('data_comb_fn', default_data_comb_fn)


        trial_filter_fn = filt
        trial_condition_fn = cond
        trial_proc_fn = proc
        data_comb_fn = comb

        # import trial_filter_functions, trial_proc_functions, trial_condition_functions
        # if isinstance(trial_filter_fn, str):
        #     trial_filter_fn = getattr(trial_filter_functions, trial_filter_fn)

        # if isinstance(trial_proc_fn, str):
        #     trial_proc_fn = getattr(trial_proc_functions, trial_proc_fn)

        # if isinstance(trial_condition_fn, str):
        #     trial_condition_fn = getattr(trial_condition_functions, trial_condition_fn)


        result = []
        error_count = 0
        for blockset in self.blocks:
            if not np.iterable(blockset):
                blockset = (blockset,)

            blockset_data = defaultdict(list)
            for te in blockset:
                if verbose:
                    print(".")

                # Filter out the trials you want
                trial_msgs = [msgs for msgs in te.trial_msgs if trial_filter_fn(te, msgs)]
                n_trials = len(trial_msgs)

                ## Call a function on each trial
                for k in range(n_trials):
                    try:
                        output = trial_proc_fn(te, trial_msgs[k], **kwargs)
                        trial_condition = trial_condition_fn(te, trial_msgs[k])
                        blockset_data[trial_condition].append(output)
                    except:
                        error_count += 1
                        print(trial_msgs[k])
                        import traceback
                        traceback.print_exc()
                        if error_count > max_errors:
                            raise Exception

            # Aggregate the data from the blockset, which may include multiple task entries
            blockset_data_comb = dict()
            for key in blockset_data:
                blockset_data_comb[key] = data_comb_fn(blockset_data[key])
            if len(list(blockset_data_comb.keys())) == 1:
                key = list(blockset_data_comb.keys())[0]
                result.append(blockset_data_comb[key])
            else:
                result.append(blockset_data_comb)

        if verbose:
            sys.stdout.write('\n')
        return result

    def proc_blocks(self, filt=None, proc=None, cond=None, comb=None, verbose=False, return_type=list, **kwargs):
        '''
        Generic framework to perform a block-level analysis on the entire dataset,
        e.g., percent of trials correct, which require analyses across trials

        Parameters
        ----------
        block_filter_fn: callable; call signature: block_filter_fn(task_entry)
            Function must return True/False values to determine if a task entry is valid for the analysis
        block_proc_fn: callable; call signature: trial_proc_fn(task_entry)
            The main workhorse function
        data_comb_fn: callable; call signature: data_comb_fn(list)
            Combine the list into the desired output structure

        Returns
        -------
        result: list
            The results of all the analysis. The length of the returned list equals len(self.blocks). Sub-blocks
            grouped by tuples are combined into a single result.
        '''
        if filt == None:
            filt = kwargs.pop('block_filter_fn', default_block_filter_fn)
        if cond == None:
            cond = kwargs.pop('block_condition_fn', default_trial_condition_fn)
        if proc == None:
            proc = kwargs.pop('block_proc_fn', default_trial_proc_fn)
        if comb == None:
            comb = kwargs.pop('data_comb_fn', default_data_comb_fn)


        block_filter_fn = filt
        block_condition_fn = cond
        block_proc_fn = proc
        data_comb_fn = comb

        # Look up functions by name, if strings are given instead of functions
        # import trial_filter_functions, trial_proc_functions, trial_condition_functions
        # if isinstance(block_filter_fn, str):
        #     block_filter_fn = getattr(trial_filter_functions, block_filter_fn)

        # if isinstance(block_proc_fn, str):
        #     block_proc_fn = getattr(block_proc_functions, block_proc_fn)

        result = []
        for blockset in self.blocks:
            if not np.iterable(blockset):
                blockset = (blockset,)

            blockset_data = []
            for te in blockset:
                if verbose:
                    print(".")

                if block_filter_fn(te):
                    blockset_data.append(block_proc_fn(te, **kwargs))

            blockset_data = data_comb_fn(blockset_data)
            result.append(blockset_data)

        if verbose:
            sys.stdout.write('\n')
        return return_type(result)

    def __repr__(self):
        if not self.name == '':
            return str(self.block_ids)
        else:
            return "TaskEntryCollection: ", self.name

    def __str__(self):
        return self.__repr__()


##################
# Query functions
##################
@mysqlreconnect
def get_task_entries_by_id(ids):
    '''
    Returns list of task entries according to the given id or multiple ids
    '''
    return list(models.TaskEntry.objects.filter(pk__in=ids))

@mysqlreconnect
def get_task_entries_by_date(startdate, enddate=datetime.date.today(), dbname='default'):
    '''
    Returns list of task entries within date range (inclusive). startdate and enddate
    are date objects. End date is optional- today's date by default.
    '''
    return list(models.TaskEntry.objects.using(dbname).filter(date__gte=startdate).filter(date__lte=enddate))

@mysqlreconnect
def get_task_entries(subj=None, date=None, task=None, dbname='default', **kwargs):
    '''
    Get all the task entries for a particular date

    Parameters
    ----------
    subj: string, optional, default=None
        Specify the beginning of the name of the subject or the full name. If not specified, blocks from all subjects are returned
    date: multiple types, optional, default=today
        Query date for blocks. The date can be specified as a datetime.date object, a tuple of (start, end) datetime.date objects,
        or a 3-tuple of ints (year, month, day).
    task: string
        name of task to filter
    kwargs: dict, optional
        Additional keyword arguments to pass to models.TaskEntry.objects.filter
    '''

    if isinstance(date, datetime.date):
        kwargs.update(dict(date__year=date.year, date__month=date.month, date__day=date.day))
    elif isinstance(date, tuple) and len(date) == 2:
        kwargs.update(dict(date__gte=date[0], date__lte=date[1]))
    elif isinstance(date, tuple) and len(date) == 3:
        kwargs.update(dict(date__year=date[0], date__month=date[1], date__day=date[2]))

    if isinstance(subj, str):
        kwargs['subject__name__startswith'] = str(subj)
    elif subj is not None:
        kwargs['subject__name'] = subj.name
    
    if isinstance(task, str):
        kwargs['task__name__startswith'] = task
    elif task is not None:
        kwargs['task__name'] = task.name

    return list(models.TaskEntry.objects.using(dbname).filter(**kwargs))

#################
# Collections
###############
def parse_blocks(blocks, cls=TaskEntry, **kwargs):
    '''
    Parse out a hierarchical structure of block ids. Used to construct TaskEntryCollection objects
    '''
    data = []
    for block in blocks:
        if np.iterable(block):
            te = parse_blocks(block, cls=cls, **kwargs)
        else:
            te = cls(block, **kwargs)
        data.append(te)
    return data

def group_ids(ids, grouping_fn=lambda te: te.calendar_date):
    '''
    Automatically group together a flat list of database IDs

    Parameters
    ----------
    ids: iterable
        iterable of ints representing the ID numbers of TaskEntry objects to group
    grouping_fn: callable, optional (default=sort by date); call signature: grouping_fn(task_entry)
        Takes a dbfn.TaskEntry as its only argument and returns a hashable and sortable object
        by which to group the ids
    '''
    keyed_ids = defaultdict(list)
    for id in ids:
        te = TaskEntry(id)
        key = grouping_fn(te)
        keyed_ids[key].append(id)

    keys = list(keyed_ids.keys())
    keys.sort()

    grouped_ids = []
    for date in keys:
        grouped_ids.append(tuple(keyed_ids[date]))
    return grouped_ids

############
# Statistics
############
def default_data_comb_fn(x):
    return x

def default_trial_filter_fn(te, trial_msgs):
    return True

def default_block_filter_fn(te):
    return True

def default_trial_proc_fn(te, trial_msgs):
    return 1

def default_trial_condition_fn(te, trial_msgs):
    return 0

def get_records_of_trained_decoders(task_entry):
    '''
    Returns unpickled decoder objects that were trained in a specified session.
    '''
    task_entry = lookup_task_entries(task_entry)
    records = models.Decoder.objects.filter(entry_id=task_entry.id)
    records = list(records)
    if len(records) == 1:
        return records[0]
    else:
        return records

def lookup_task_entries(*task_entry, dbname='default'):
    '''
    Enable multiple ways to specify a task entry, e.g. by primary key or by
    object
    '''
    if len(task_entry) == 1:
        task_entry = task_entry[0]
        if isinstance(task_entry, models.TaskEntry):
            pass
        elif isinstance(task_entry, int):
            task_entry = models.TaskEntry.objects.using(dbname).get(pk=task_entry) #get_task_entry(task_entry)
        return task_entry
    else:
        return [lookup_task_entries(x) for x in task_entry]

def get_task_id(name, dbname='default'):
    '''
    Returns the task ID for the specified task name.
    '''
    return models.Task.objects.using(dbname).get(name=name).pk

def get_decoder_entry(entry, dbname='default'):
    '''Returns the database entry for the decoder used in the session. Argument can be a task entry
    or the ID number of the decoder entry itself.
    '''
    if isinstance(entry, int):
        return models.Decoder.objects.using(dbname).get(pk=entry)
    else:
        params = json.loads(entry.params)
        if 'decoder' in params:
            return models.Decoder.objects.using(dbname).get(pk=params['decoder'])
        elif 'bmi' in params:
            return models.Decoder.objects.using(dbname).get(pk=params['bmi'])
        else:
            return None

def get_decoder_name(entry, dbname='default'):
    '''
    Returns the filename of the decoder used in the session.
    Takes TaskEntry object.
    '''
    entry = lookup_task_entries(entry)
    try:
        decid = json.loads(entry.params)['decoder']
    except:
        decid = json.loads(entry.params)['bmi']
    return models.Decoder.objects.using(dbname).get(pk=decid).path

def get_decoder_name_full(entry, dbname='default'):
    entry = lookup_task_entries(entry)
    decoder_basename = get_decoder_name(entry)
    sys_path = models.System.objects.using(dbname).get(name='bmi')
    return os.path.join(sys_path, decoder_basename)

def get_decoder(entry, dbname='default'):
    entry = lookup_task_entries(entry)
    filename = get_decoder_name_full(entry)
    dec = pickle.load(open(filename, 'r'))
    dec.db_entry = get_decoder_entry(entry, dbname=dbname)
    dec.name = dec.db_entry.name
    return dec

def get_params(entry):
    '''
    Returns a dict of all task params for session.
    Takes TaskEntry object.
    '''
    return json.loads(entry.params)

def get_param(entry,paramname):
    '''
    Returns parameter value.
    Takes TaskEntry object.
    '''
    return json.loads(entry.params)[paramname]

def get_task_name(entry, dbname='default'):
    '''
    Returns name of task used for session.
    Takes TaskEntry object.
    '''
    return models.Task.objects.using(dbname).get(pk=entry.task_id).name

def get_date(entry):
    '''
    Returns date and time of session (as a datetime object).
    Takes TaskEntry object.
    '''
    return entry.date

def get_notes(entry):
    '''
    Returns notes for session.
    Takes TaskEntry object.
    '''
    return entry.notes

def get_subject(entry, dbname='default'):
    '''
    Returns name of subject for session.
    Takes TaskEntry object.
    '''
    return models.Subject.objects.using(dbname).get(pk=entry.subject_id).name

def get_length(entry):
    '''
    Returns length of session in seconds.
    Takes TaskEntry object.
    '''
    try:
        report = json.loads(entry.report)
    except:
        return 0.0
    return report[-1][2]-report[0][2]

def get_success_rate(entry):
    '''
    Returns (# of trials rewarded)/(# of trials intiated).
    Takes TaskEntry object.
    '''
    try:
        report = json.loads(entry.report)
    except: return 0.0
    total=0.0
    rew=0.0
    for s in report:
        if s[0]=='reward':
            rew+=1
            total+=1
        if s[0]=='hold_penalty' or s[0]=='timeout_penalty':
            total+=1
    return rew/total

def get_completed_trials(entry):
    '''
    Returns # of trials rewarded
    '''
    try:
        report = json.loads(entry.report)
    except: return 0.0
    return len([s for s in report if s[0]=="reward"])

def get_initiate_rate(entry):
    '''
    Returns average # of trials initated per minute.
    Takes TaskEntry object.
    '''
    length = get_length(entry)
    try:
        report = json.loads(entry.report)
    except: return 0.0
    count=0.0
    for s in report:
        if s[0]=='reward' or s[0]=='hold_penalty' or s[0]=='timeout_penalty':
            count+=1
    return count/(length/60.0)

def get_reward_rate(entry):
    '''
    Returns average # of trials completed per minute.
    Takes TaskEntry object.
    '''
    try:
        report = json.loads(entry.report)
    except: return 0.0
    count=0.0
    rewardtimes = []
    for s in report:
        if s[0]=='reward':
            count+=1
            rewardtimes.append(s[2])
    if len(rewardtimes)==0:
        return 0
    else:
        length = rewardtimes[-1] - report[0][2]
        return count/(length/60.0)

def session_summary(entry):
    '''
    Prints a summary of info about session.
    Takes TaskEntry object.
    '''
    entry = lookup_task_entries(entry)
    print("Subject: ", get_subject(entry))
    print("Task: ", get_task_name(entry))
    print("Date: ", str(get_date(entry)))
    hours = np.floor(get_length(entry)/3600)
    mins = np.floor(get_length(entry)/60) - hours*60
    secs = get_length(entry) - mins*60
    print("Length: " + str(int(hours))+ ":" + str(int(mins)) + ":" + str(int(secs)))
    try:
        print("Assist level: ", get_param(entry,'assist_level'))
    except:
        print("Assist level: 0")
    print("Completed trials: ", get_completed_trials(entry))
    print("Success rate: ", get_success_rate(entry)*100, "%")
    print("Reward rate: ", get_reward_rate(entry), "trials/minute")
    print("Initiate rate: ", get_initiate_rate(entry), "trials/minute")

###############
# Decoder query
###############
def get_decoder_parent(decoder, dbname='default'):
    '''
    decoder = database record of decoder object
    '''
    entryid = decoder.entry_id
    te = get_task_entries_by_id(entryid, dbname=dbname)[0]
    return get_decoder_entry(te)

def get_decoder_sequence(decoder, dbname='default'):
    '''
    decoder = database record of decoder object
    '''
    parent = get_decoder_parent(decoder, dbname=dbname)
    if parent is None:
        return [decoder]
    else:
        return [decoder] + get_decoder_sequence(parent, dbname=dbname)

def search_by_decoder(decoder, dbname='default'):
    '''
    Returns task entries that used specified decoder. Decoder argument can be
    decoder entry or entry ID.
    '''

    if isinstance(decoder, int):
        decid = decoder
    else:
        decid = decoder.id
    blocks = list(models.TaskEntry.objects.using(dbname).filter(params__contains='"bmi": '+str(decid))) + list(models.TaskEntry.objects.using(dbname).filter(params__contains='"decoder": '+str(decid)))
    return blocks

def search_by_units(unitlist, decoderlist=None, exact=False, dbname='default'):
    '''
    Returns decoder entries that contain the specified units. If exact is True,
    returns only decoders whose unit lists match unitlist exactly, otherwise
    returns decoders that contain units in unitlist in addition to others. If
    given a list of decoder entries, only searches within those, otherwise searches
    all decoders in database.
    '''
    if decoderlist is not None:
        all_decoders = decoderlist
    else:
        all_decoders = models.Decoder.objects.using(dbname).all()
    subset = set(tuple(unit) for unit in unitlist)
    dec_list = []
    for dec in all_decoders:
        try:
            decobj = dec.get()
            decset = set(tuple(unit) for unit in decobj.units)
            if subset==decset:
                dec_list = dec_list + [dec]
            elif not exact and subset.issubset(decset):
                dec_list = dec_list + [dec]
        except:
            pass
    return dec_list

def load_last_decoder():
    '''
    Returns the decoder object corresponding to the last decoder trained and
    added to the database
    '''
    all_decoder_records = models.Decoder.objects.all()
    record = all_decoder_records[len(all_decoder_records)-1]
    return record.load()


######################
## Filter functions
######################
def _assist_level_0(task_entry_model):
    te = TaskEntry(task_entry_model)
    if 'assist_level' in te.params:
        assist_level = te.params['assist_level']
        return np.all(np.array(assist_level) == 0)
    else:
        return True

def using_decoder(decoder_type):
    from tasks import performance
    def _check_decoder(task_entry_model):
        te = performance._get_te(task_entry_model)
        try:
            return (decoder_type in str(te.decoder_type))
        except:
            return False
    return _check_decoder

def min_trials(min_trial_count):
    from tasks import performance
    def fn(task_entry_model):
        te = performance._get_te(task_entry_model)
        try:
            return te.n_trials >= min_trial_count
        except:
            return False
    return fn

##########
## Filters
##########
def get_blocks_after(id, **kwargs):
    return [x for x in models.TaskEntry.objects.filter(**kwargs) if x.id > id]

def get_bmi_blocks(date, subj='C'):
    blocks = TaskEntrySet.get_blocks(filter_fns=[min_trials(50)], subj=subj, date=date, task__name__startswith='bmi') + TaskEntrySet.get_blocks(filter_fns=[min_trials(5)], subj=subj, date=date, task__name__startswith='clda')
    blocks.sort(key=lambda x: x.date)
    return blocks

def bmi_filt(te):
    bmi_feat = models.Feature.objects.using(te._state.db).get(name='bmi')
    return bmi_feat in te.feats.all()

##################
## aolab functions
##################
def get_entry_details(entry):
    '''
    Returns subject, id, and date for the given entry
    '''
    subject = get_subject(entry)
    te = entry.id
    date = get_date(entry).date()
    return subject, te, date

def get_sessions(subject, date, task_name, session_name=None, exclude_ids=[], filter_fn=lambda x:True, dbname='default'):
    '''
    Returns list of subject, date, and id for all sessions on the given date
    '''
    if session_name:
        entries = get_task_entries(subj=subject, task=task_name, date=date, session=session_name, dbname=dbname)
    else:
        entries = get_task_entries(subj=subject, task=task_name, date=date, dbname=dbname)
    if len(entries) == 0:
        import warnings
        warnings.warn("No entries found")
        return [], [], []
    subject, te, date = zip(*[get_entry_details(e) for e in entries if filter_fn(e) and e.id not in exclude_ids])
    return subject, te, date
    
def get_mc_sessions(subject, date, mc_task_name='manual control', dbname='default', **kwargs):
    '''
    Returns list of subject, date, and id for all bmi control sessions on the given date
    '''
    return get_sessions(subject, date, task_name=mc_task_name, dbname=dbname, **kwargs)
    
def get_bmi_sessions(subject, date, session_name='training', bmi_task_name='bmi control', dbname='default', **kwargs):
    '''
    Returns list of subject, date, and id for all bmi control sessions on the given date
    '''
    return get_sessions(subject, date, session_name=session_name, task_name=bmi_task_name, dbname=dbname, **kwargs)


def get_preprocessed_sources(entry):
    '''
    Returns a list of datasource names that should be preprocessed for the given entry
    '''
    params = entry.task_params
    sources = ['exp']
    if 'record_headstage' in params and params['record_headstage']:
        sources.append('broadband')
        sources.append('lfp')
    sources.append('eye')
    
    return sources
    

def get_rawfiles_for_taskentry(e, system_subfolders=None):
    '''
    Gets the hdf and ecube files associated with each task entry 
    
    Args:
        entry (TaskEntry): recording to find raw files for 
        system_subfolders (dict, optional): dictionary of system subfolders where the 
           data for that system is located. If None, defaults to the system name
    
    Returns: 
        files : list of (system, filepath) for each datafile associated with this task entry 
    '''
    file_list = e.get_data_files(dbname=e._state.db)
    files = {}
    for df in file_list:
        if df.system.name == 'optitrack' or df.system.name == 'e3v':
            path = os.path.normpath(df.path)
            parts = path.split(os.sep)
            if len(parts) > 1:
                filepath = os.path.join(*parts[1:])
            filepath = filepath.replace(':', '_')
        else:
            filepath = os.path.basename(df.path)
        if system_subfolders is None:
            files[df.system.name] = os.path.join(df.system.name, filepath)
        elif df.system.name in system_subfolders:
            files[df.system.name] = os.path.join(system_subfolders[df.system.name], filepath)
        else:
            files[df.system.name] = filepath

    return files



class SubclassDict(dict):
    '''
    A special dict that returns the associated Django database model 
    if the queried item is a subclass of any of the keys
    '''
    def __getitem__(self, name):
        try:
            return super(self.__class__, self).__getitem__(name)
        except KeyError:
            for inst, model in list(self.items()):
                if issubclass(name, inst):
                    return model
        raise KeyError


import time
from riglib import source
from riglib.quattrocento import EMG, comms
from riglib.bmi import state_space_models, train, extractor
import numpy as np
import unittest

STREAMING_DURATION = 3

class TestStreaming(unittest.TestCase):

    @unittest.skip('works')
    def test_direct(self):
        qt = comms.QuattroOtlight(host='128.95.215.191', refresh_freq=32)
        qt.setup()
        data = qt.read_emg()
        for d in data:
            print(d.shape)
        qt.tear_down()

    @unittest.skip('works')
    def test_quattrocento_stream(self):
        channels = [1, 3]
        bb = EMG(channels=channels) # Note: this is not how this is normally used, just testing
        bb.start()
        data = bb.qt.read_emg()
        bb.stop()
        print(data[2].shape)
    
    @unittest.skip('works')
    def test_broadband_class(self):
        channels = [1, 3]
        bb = EMG(channels=channels)
        bb.start()
        ch_data = []
        for d in range(2):
            for i in range(len(channels)):
                ch, data = bb.get()
                ch_data.append(data)
                print(f"Got channel {ch} with {data.shape} samples")
        bb.stop()

    @unittest.skip('works')
    def test_broadband_datasource(self):
        channels = [1, 62]
        ds = source.MultiChanDataSource(EMG, channels=channels)
        ds.start()
        time.sleep(STREAMING_DURATION)
        data = ds.get_new(channels)
        ds.stop()
        data = np.array(data)

        n_samples = int(EMG.update_freq * STREAMING_DURATION)

        self.assertEqual(data.shape[0], len(channels))
        self.assertEqual(data.shape[1], n_samples)

    @unittest.skip('works')
    def test_update_frequency(self):
        channels = [1, 62]
        ds = source.MultiChanDataSource(EMG, channels=channels, send_data_to_sink_manager=True)
        ds.start()
        t0 = time.perf_counter()
        n_packets = 0
        while time.perf_counter() - t0 < STREAMING_DURATION:
            data = ds.get_new(channels)
            if len(data[0]) > 0 or len(data[1]) > 0:
                n_packets += 1
            time.sleep(0.001)
        ds.stop()

        expected_packets = 32 * int(STREAMING_DURATION)

        self.assertAlmostEqual(n_packets, expected_packets, delta=10)

    @unittest.skip('works')
    def test_ds_with_extractor(self):
        # Create the datasource
        channels = [1, 62]
        ds = source.MultiChanDataSource(EMG, channels=channels)

        # Make a feature extractor
        extr = extractor.LFPMTMPowerExtractor(ds, channels=channels, bands=[(90,110)], win_len=0.1, fs=2048)

        # Run the feature extractor
        extract_rate = 1/60
        feats = []
        ds.start()
        t0 = time.perf_counter()
        while (time.perf_counter() - STREAMING_DURATION < t0):
            neural_features_dict = extr(time.perf_counter())
            feats.append(neural_features_dict['lfp_power'])
            time.sleep(extract_rate)
        
        ds.stop()
        data = np.array(feats)
        print(feats)
        print(data.shape)
        self.assertEqual(data.shape[1], len(channels))
        self.assertEqual(data.shape[0], STREAMING_DURATION/extract_rate)

if __name__ == '__main__':
    unittest.main()


'''
Base code for 'saveHDF' feature in experiments for writing data to an HDF file during experiment
'''

import tables
import numpy as np

compfilt = tables.Filters(complevel=5, complib="zlib", shuffle=True)

class MsgTable(tables.IsDescription):
    '''
    Pytables custom table atom type used for the HDF tables named *_msgs
    '''
    time = tables.UIntCol()
    msg = tables.StringCol(256)

class HDFWriter(object):
    ''' 
    Used by the SaveHDF feature (features.hdf_features.SaveHDF) to save data 
    to an HDF file in "real-time", as the task is running
    '''
    def __init__(self, filename):
        '''
        Constructor for HDFWriter

        Parameters
        ----------
        filename : string
            Name of file to use to send data

        Returns
        -------
        HDFWriter instance
        '''
        print("HDFWriter: Saving datafile to %s"%filename)
        self.h5 = tables.open_file(filename, "w")
        self.data = {}
        self.msgs = {}
        self.f = []
    
    def register(self, name, dtype, include_msgs=True):
        '''
        Create a table in the HDF file corresponding to the specified source name and data type

        Parameters
        ----------
        system : string
            Name of the system being registered
        dtype : np.dtype instance
            Datatype of incoming data, for later decoding of the binary data during analysis
        include_msgs : boolean, optional, default=True
            Flag to indicated whether a table should be created for "msgs" from the current source (default True)

        Returns
        -------
        None
        '''
        print("HDFWriter registered %r" % name)
        print(dtype)
        if dtype.subdtype is not None:
            #just a simple dtype with a shape
            dtype, sliceshape = dtype.subdtype
            arr = self.h5.create_earray("/", name, tables.Atom.from_dtype(dtype), 
                shape=(0,)+sliceshape, filters=compfilt)
        else:
            arr = self.h5.create_table("/", name, dtype, filters=compfilt)

        self.data[name] = arr
        if include_msgs:
            msg = self.h5.create_table("/", name+"_msgs", MsgTable, filters=compfilt)
            self.msgs[name] = msg
    
    def send(self, system, data):
        '''
        Add a new row to the HDF table for 'system' and fill it with the 'data' values

        Parameters
        ----------
        system : string
            Name of system where the data originated
        data : object
            Data to send. Must have a '.tostring()' attribute

        Returns
        -------
        None
        '''
        if system in self.data:
            if data is not None:
                self.data[system].append(data)

    def sendMsg(self, msg):
        '''
        Write a string to the *_msgs table for each system registered with the HDF sink

        Parameters
        ----------
        msg : string
            Message to link to the current row of the HDF table

        Returns
        -------
        None
        '''  
        for system in list(self.msgs.keys()):
            row = self.msgs[system].row
            row['time'] = len(self.data[system])
            row['msg'] = msg
            row.append()

    def sendAttr(self, system, attr, value):
        '''
        While the HDF writer process is running, set an attribute of the table
        (not sure that this has ever been tested..)

        Parameters
        ----------
        system : string
            Name of the table where the attribute should be set
        attr : string 
            Name of the attribute
        value : object
            Value of the attribute to set

        Returns
        -------
        None
        '''
        if system in self.data:
            self.data[system].attrs[attr] = value
    
    def close(self):
        '''
        Close the HDF file so that it saves properly after the process terminates
        '''
        self.h5.close()
        print("Closed hdf")




import setuptools

setuptools.setup(
    name="aolab-bmi3d",
    version="1.2.3",
    author="Lots of people",
    description="electrophysiology experimental rig library",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={
        "riglib": ["*.wav"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "django==4.1",
        "celery",
        "jinja2",
        "scipy",
        "numpy==1.26",
        "traits",
        "pandas",
        "patsy",
        "statsmodels",
        "pygame==2.5.0",
        "PyOpenGL",
        "pylibftdi",
        "sphinx",
        "numpydoc",
        "tornado",
        "tables",
        "h5py",
        "pymysql",
        "matplotlib",
        "pyfirmata",
        "hdfwriter==0.1.4",
    ]
)



# Generated by Django 3.1.4 on 2021-05-16 15:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0007_taskentry_template'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskentry',
            name='metadata',
            field=models.TextField(default=''),
        ),
    ]



#SerialLink.ploT Graphical display and animation

# This is a dirrect port from the matlab robot toolbox.  it is still a
# work in progress i was working on tgetting the plot options to work  
# correctly.  i was going to use the same method to 3d plot that was
# used in trplot.  there is still alot of work to be done to get this
# function up and running.
# Ryan James Steindl

#
# R.plot(Q, options) displays a graphical animation of a robot based on 
# the kinematic model.  A stick figure polyline joins the origins of
# the link coordinate frames. The robot is displayed at the joint angle Q, or 
# if a matrix it is animated as the robot moves along the trajectory.
#
# The graphical robot object holds a copy of the robot object and
# the graphical element is tagged with the robot's name (.name property).
# This state also holds the last joint configuration which can be retrieved,
# see PLOT(robot) below.
#
# Figure behaviour::
# If no robot of this name is currently displayed then a robot will
# be drawn in the current figure.  If hold is enabled (hold on) then the
# robot will be added to the current figure.
#
# If the robot already exists then that graphical model will be found 
# and moved.
#
# Multiple views of the same robot::
#
# If one or more plots of this robot already exist then these will all
# be moved according to the argument Q.  All robots in all windows with 
# the same name will be moved.
#
# Multiple robots in the same figure::
#
# Multiple robots can be displayed in the same plot, by using "hold on"
# before calls to plot(robot).  
#
# Graphical robot state::
#
# The configuration of the robot as displayed is stored in the SerialLink object
# and can be accessed by the read only object property R.q.
#
# Graphical annotations and options::
#
# The robot is displayed as a basic stick figure robot with annotations 
# such as:
# - shadow on the floor
# - XYZ wrist axes and labels
# - joint cylinders and axes
# which are controlled by options.
#
# The size of the annotations is determined using a simple heuristic from 
# the workspace dimensions.  This dimension can be changed by setting the 
# multiplicative scale factor using the 'mag' option.
#
# Options::
#  workspace = W          size of robot 3D workspace, W = [xmn, xmx ymn ymx zmn zmx]
#  delay = d              delay betwen frames for animation (s)
#  cylinder = C           color for joint cylinders, C=[r g b]
#  mag = scale            annotation scale factor
#  perspective = True     type of camera view False for'ortho'
#  raise = Ture           controls autoraise of current figure on plot false for 'noraise'
#  render = Ture          controls shaded rendering after drawing false for 'norender'
#  loop = Ture            controls endless loop mode false for 'noloop'
#  base = Ture            controls display of base false for 'nobase' 
#  wrist = Ture           controls display of wrist false for 'nowrist' 
#  shadow = Ture          controls display of shadow false for 'noshadow'
#  name = Ture            display the robot's name false for'noname'
#  xyz = Ture             wrist axis label false for 'noa'      
#  jaxes = Ture           control display of joint axes false for 'nojaxes'   
#  joints = Ture          controls display of joints false for 'nojoints' 
#
# 
# See also plotbotopt, SerialLink.fkine.


# HANDLES:
#
#  A robot comprises a bunch of individual graphical elements and these are 
# kept in a structure which can be stored within the .handle element of a
# robot object:
#   h.robot     the robot stick figure
#   h.shadow    the robot's shadow
#   h.x     wrist vectors
#   h.y
#   h.z
#   h.xt        wrist vector labels
#   h.yt
#   h.zt
#
#  The plot function returns a new robot object with the handle element set.
#
# For the h.robot object we additionally: 
#   - save this new robot object as its UserData
#   - tag it with the name field from the robot object
#
#  This enables us to find all robots with a given name, in all figures,
# and update them.



# Copyright (C) 1993-2011, by Peter I. Corke
#
# This file is part of The Robotics Toolbox for Matlab (RTB).
# 
# RTB is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# RTB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Leser General Public License
# along with RTB.  If not, see <http://www.gnu.org/licenses/>.
#
# http://www.petercorke.com

from numpy import *
from .SerialLink import *
from .Link import * 
from .utility import *
from .transform import *
import matplotlib.pyplot as plt
import numpy as np


def plot(robot, tg, **kwargs): 
    ## set all the defaults for the kwargs
    defaults = dict(workspace=None, delay=None, cylinder=None, mag=None, perspective=True, Raise=True, render=True, loop=True, base=True, wrist=True, shadow=True, name=True, xyz=True, jaxes=True, joints=True)
    for item, val in defaults.items():
        if item not in kwargs:
            kwargs[item] = val

    opt = struct()
    opt.__dict__ = kwargs

    # opts = PLOT(robot, options)
    #  just convert options list to an options struct
    
    #opt = plot_options(robot, workspace, delay, cylinder, mag, perspective, Raise, render, loop, base, wrist, shadow, name, xyz, jaxes, joints)

    #Check to see if correct numper of joint angles in Q 
    stuff = tg.shape[0] #np.numrows(tg)
    n = robot.n

    if len(tg) != robot.n:
        error('Insufficient columns in q')

    
    # get handle of any existing robot of same name
    rh = findobj('Tag', robot.name)

    if isempty(rh) or isempty( get(gcf, 'Children')):
        # no robot of this name exists

        # create one
        ax = plt.subplot(111)
        h = create_new_robot(robot, opt)

        # save the handles in the passed robot object, and
        # attach it to the robot as user data.
        robot.handle = h
        set(h.robot, 'Tag', robot.name)
        set(h.robot, 'UserData', robot)

        rh = h.robot
    

    if ishold and isempty( findobj(gca, 'Tag', robot.name)):
        # if hold is on and no robot of this name in current axes
        h = create_new_robot(robot, opt)
        # save the handles in the passed robot object, and
        # attach it to the robot as user data.
        robot.handle = h
        set(h.robot, 'Tag', robot.name)
        set(h.robot, 'UserData', robot)

        rh = h.robot
    
    if Raise:
        figure(gcf)
    

    while True:
        for p in range(0,stuff):
            for r in rh.T:
                animate( get(r, 'UserData'), tg[p,:], opt)
                if opt.delay > 0:
                    pause(opt.delay)
        
        if ~opt.loop:
            break

    # save the last joint angles away in the graphical robot
    for r in rh.T:
        rr = get(r, 'UserData')
        rr.q = tg[-1,:]
        set(r, 'UserData', rr)
    
    return robot
    

#PLOT_OPTIONS
#
#   o = PLOT_OPTIONS(robot, options)
#
# Returns an options structure

class struct(object):
    pass

def plot_options(robot, workspace, delay, cylinder, mag, perspective, Raise, render, loop, base, wrist, shadow, name, xyz, jaxes, joints):
    '''
    # process a cell array of options and return a struct

    # define all possible options and their default values
    '''
    o = struct()
    o.erasemode = 'normal'

    # set workspace
    if workspace == None:
        #
        # simple heuristic to figure the maximum reach of the robot
        #
        L = robot.links
        reach = 0
        #for i in range(1, robot.n):
        for link in robot.links:
            reach = reach + abs(link.a) + abs(link.d)

        o.workspace = [-reach, reach, -reach, reach, -reach, reach]
        o.mag = reach/10
    else:
        reach = min(abs(workspace))

    #set mag
    o.mag = o.magscale * reach/10

    #set delay
    if delay is not None:
        o.delay = delay
    else:
        delay = 0.1

    # set cylinder
    if cylinder is not None:
        o.cylinder = cylinder
    else:
        o.cylinder = [0, 0, 0.7]

    #set perspective
    if perspective == True:
        o.perspective = True
    else:
        o.perspective = False

    #set Raise
    if Raise:
        o.Raise = True
    else:
        o.Raise = False

    #set render
    if render:
        o.render = True
    else:
        o.render = False

    #Set loop
    if loop:
        o.loop = True
    else:
        o.loop = False

    #set base
    if base == True:
        o.base = True
    else:
        o.base = False
    
    #set wrist
    if wrist == True:
        o.wrist = True
    else:
        o.wrist = False
        
    #set shadow
    if shadow == True:
        o.shadow = True
    else:
        o.shadow = False

    #set name
    if name == True:
        o.name = True
    else:
        o.name = False

    #set xyz
    if  xyz == True:
        o.xyz = 'xyz'
    else:
        o.xyz = 'noa'
    
    #set jaxes
    if jaxes == true:
        o.jaxes = True
    else:
        o.jaxes = False

    #set joints
    if joints:
        o.joints = True
    else:
        o.joints = False

    #############
    #finish function by outputing the plot options object.
    return o

 

#CREATE_NEW_ROBOT
# 
#   h = CREATE_NEW_ROBOT(robot, opt)
#
# Using data from robot object and options create a graphical robot in
# the current figure.
#
# Returns a structure of handles to graphical objects.
#
# If current figure is empty, draw robot in it
# If current figure has hold on, add robot to it
# Otherwise, create new figure and draw robot in it.
#   

# h.mag
# h.zmin
# h.robot   the line segment that is the robot
# h.shadow  the robot's shadow
# h.x       the line segment that is T6 x-axis
# h.y       the line segment that is T6 x-axis
# h.z       the line segment that is T6 x-axis
# h.xt      text for T6 x-axis
# h.yt      text for T6 y-axis
# h.zt      text for T6 z-axis
# h.joint(i)        the joint i cylinder or box
# h.jointaxis(i)    the line segment that is joint i axis
# h.jointlabel(i)   text for joint i label

def create_new_robot(robot, opt):
    h.mag = opt.mag

    #
    # setup an axis in which to animate the robot
    #
    # handles not provided, create graphics
    #disp('in creat_new_robot')
    if ~ishold:
        # if current figure has hold on, then draw robot here
        # otherwise, create a new figure
        axis(opt.workspace)
    
    xlabel('X')
    ylabel('Y')
    zlabel('Z')
    set(gca, 'drawmode', 'fast')
    plt.grid(True)


    zlim = get(gca, 'ZLim')
    h.zmin = zlim(1)

    if opt.base:
        b = transl(robot.base)
        line('xdata', np.r_[b(1), b(1)], 'ydata', np.r_[b(2), b(2)], 'zdata', np.r_[h.zmin, b(3)], 'LineWidth', 4, 'color', 'red')
    
    
    if opt.name:
        b = transl(robot.base)
        text(b(1), b(2)-opt.mag, [' ', robot.name], 'FontAngle', 'italic', 'FontWeight', 'bold')
    
    # create a line which we will
    # subsequently modify.  Set erase mode to xor for fast
    # update
    #
    h.robot = line(**robot.lineopt)
    
    if opt.shadow:
        h.shadow = line(robot.shadowopt, 'Erasemode', opt.erasemode)
    

    if opt.wrist:   
        h.x = line('xdata', np.r_[0,0], 'ydata', np.r_[0,0], 'zdata', np.r_[0,0], 'color', 'red')
        h.y = line('xdata', np.r_[0,0], 'ydata', np.r_[0,0], 'zdata', np.r_[0,0], 'color', 'green')
        h.z = line('xdata', np.r_[0,0], 'ydata', np.r_[0,0], 'zdata', np.r_[0,0], 'color', 'blue')
        h.xt = text(0, 0, opt.wristlabel(1), 'FontWeight', 'bold', 'HorizontalAlignment', 'Center')
        h.yt = text(0, 0, opt.wristlabel(2), 'FontWeight', 'bold', 'HorizontalAlignment', 'Center')
        h.zt = text(0, 0, opt.wristlabel(3), 'FontWeight', 'bold', 'HorizontalAlignment', 'Center')

    

    #
    # display cylinders (revolute) or boxes (pristmatic) at
    # each joint, as well as axis centerline.
    #
    L = robot.links
    for i in range(1, robot.n):
        
        if opt.joints:

            # cylinder or box to represent the joint
            if L(i).sigma == 0:
                N = 16
            else:
                N = 4
            
            # define the vertices of the cylinder
            [xc,yc,zc] = cylinder(opt.mag/4, N)
            zc[zc==0] = -opt.mag/2
            zc[zc==1] = opt.mag/2

            # create vertex color data
            cdata = zeros(size(xc))
            for j in range(0, 2):
                cdata[:,:,j] = opt.cylinder[j]
            
            # render the surface
            h.joint[i] = surface(xc,yc,zc,cdata)
            
            # set the surfaces to be smoothed and translucent
            set(h.joint(i), 'FaceColor', 'interp')
            set(h.joint(i), 'EdgeColor', 'none')
            set(h.joint(i), 'FaceAlpha', 0.7)

            # build a matrix of coordinates so we
            # can transform the cylinder in animate()
            # and hang it off the cylinder
            xyz = np.vstack([xc.reshape(-1,1), yc.reshape(-1,1), zc.reshape(-1,1), np.ones(1,2*N+2)])
            set(h.joint(i), 'UserData', xyz)
        

        if opt.jaxes:
            # add a dashed line along the axis
            h.jointaxis[i] = line('xdata', [0,0], 'ydata', [0,0], 'zdata', [0,0], 'color', 'blue', 'linestyle', ':')
            h.jointlabel[i] = text(0, 0, 0, num2str(i), 'HorizontalAlignment', 'Center')
        
    

#ANIMATE   move an existing graphical robot
#
#   animate(robot, q)
#
# Move the graphical robot to the pose specified by the joint coordinates q.
# Graphics are defined by the handle structure robot.handle.

def animate(robot, q, opt):

    n = robot.n
    h = robot.handle
    L = robot.links

    mag = h.mag

    b = transl(robot.base)
    x = b(1)
    y = b(2)
    z = b(3)

    xs = b(1)
    ys = b(2)
    zs = h.zmin

    # compute the link transforms, and record the origin of each frame
    # for the animation.
    t = robot.base
    Tn = t
    for j in range(n):
        Tn[:,:,j] = t

        t = t * L(j).A(q(j))

        x = np.r_[x, t(1,4)]
        y = np.r_[y, t(2,4)]
        z = np.r_[z, t(3,4)]
        xs = np.r_[xs, t(1,4)]
        ys = np.r_[ys, t(2,4)]
        zs = np.r_[zs, h.zmin]
    
    t = t *robot.tool

    #
    # draw the robot stick figure and the shadow
    #
    set(h.robot,'xdata', x, 'ydata', y, 'zdata', z)
    if isfield(h, 'shadow'):
        set(h.shadow,'xdata', xs, 'ydata', ys, 'zdata', zs)
    
    

    #
    # display the joints as cylinders with rotation axes
    #
    if isfield(h, 'joint'):
        xyz_line = np.array([[0, 0],
                             [0, 0], 
                             [-2*mag, 2*mag], 
                             [1, 1]])

        for j in range(n+1):
            # get coordinate data from the cylinder
            xyz = get(h.joint(j), 'UserData')
            xyz = Tn[:,:,j] * xyz
            ncols = numcols(xyz)/2
            xc = reshape(xyz[1,:], 2, ncols)
            yc = reshape(xyz[2,:], 2, ncols)
            zc = reshape(xyz[3,:], 2, ncols)

            set(h.joint(j), 'Xdata', xc, 'Ydata', yc, 'Zdata', zc)

            xyzl = Tn[:,:,j] * xyz_line
                   
            if isfield(h, 'jointaxis'):
                set(h.jointaxis(j), 'Xdata', xyzl[1,:], 'Ydata', xyzl[2,:], 'Zdata', xyzl[3,:])
                set(h.jointlabel(j), 'Position', xyzl[1:3,1])
            

    #
    # display the wrist axes and labels
    #
    if isfield(h, 'x'):
        #
        # compute the wrist axes, based on final link transformation
        # plus the tool transformation.
        #
        xv = t*[mag,0,0,1]
        yv = t*[0,mag,0,1]
        zv = t*[0,0,mag,1]

        #
        # update the line segments, wrist axis and links
        #
        set(h.x,'xdata',[t(1,4), xv(1)], 'ydata', [t(2,4), xv(2)], 'zdata', [t(3,4), xv(3)])
        set(h.y,'xdata',[t(1,4), yv(1)], 'ydata', [t(2,4), yv(2)], 'zdata', [t(3,4), yv(3)])
        set(h.z,'xdata',[t(1,4), zv(1)], 'ydata', [t(2,4), zv(2)], 'zdata', [t(3,4), zv(3)])
        set(h.xt, 'Position', xv[1:3])
        set(h.yt, 'Position', yv[1:3])
        set(h.zt, 'Position', zv[1:3])
    



import os
import sys
import time
import pygame

os.environ['SDL_VIDEO_WINDOW_POS'] = "1024,0"
os.environ['SDL_VIDEO_X11_WMCLASS'] = "monkey_experiment"
pygame.init()

flags = pygame.DOUBLEBUF | pygame.HWSURFACE | pygame.NOFRAME
pygame.display.set_mode((3840,1080), flags)

surf = pygame.display.get_surface()
'''
coords = map(int, sys.argv[1:])
c1 = coords[:]
c1[2] = coords[2] - coords[0]
c1[3] = coords[3] - coords[1]
c2 = c1[:]
c2[0] += 1920

while True:
	surf.fill((0,0,0))
	surf.fill((255,255,255), rect=c1)
	surf.fill((255,255,255), rect=c2)
	pygame.display.flip()
'''
while True:
    surf.fill((0,0,0))
    surf.set_at((960, 540), (255,255,255))
    surf.set_at((2880, 540), (255,255,255))
    for evt in pygame.event.get(pygame.MOUSEBUTTONDOWN):
        print(evt)
    pygame.display.flip()
    time.sleep(0.01)



import serial
import pyfirmata, time
import serial.tools.list_ports
import numpy as np
from riglib.gpio import ArduinoGPIO
from features.peripheral_device_features import ForceSensorControl

# The following command can be used for finding available ports
# python -m serial.tools.list_ports

import unittest


class ForceSensorTests(unittest.TestCase):

    @unittest.skip("")
    def test_gpio(self):
        gpio = ArduinoGPIO('/dev/forcesensor', enable_analog=True)
        t0 = time.time()
        while time.time() - t0 < 5:
            print(gpio.analog_read(0))
            time.sleep(0.1)
        
    def test_source(self):
        f = ForceSensorControl()
        f.init()
        t0 = time.time()
        while time.time() - t0 < 5:
            print(f.force_sensor.get())
            time.sleep(0.1)


if __name__ == '__main__':
    unittest.main()


'''
Code for getting data from kinarm
'''
import numpy as np
from .source import DataSourceSystem
import serial

class TouchData(DataSourceSystem):
    '''
    Client for data streamed from kinarm, compatible with riglib.source.DataSource
    '''
    update_freq = 1000.

    #  dtype is the numpy data type of items that will go 
    #  into the (multi-channel, in this case) datasource's ringbuffer
    dtype = np.dtype((float, (5,5)))

    def __init__(self):
        '''
            def __init__(self, addr=("192.168.0.8", 9090)):
        
        Constructor for Kinarmdata and connect to server

        Parameters
        ----------
        addr : tuple of length 2
            (client (self) IP address, client UDP port)

        
        self.conn = kinarmsocket.KinarmSocket(addr)
        self.conn.connect()
        '''
        self.conn = TouchDataInterface()
        self.conn.connect()

    def start(self):
        '''
        Start receiving data
        '''
        self.data = self.get_iterator()


    def get_iterator(self):
        assert self.conn.connected, "Socket is not connected, cannot get data"
        while self.conn.connected:
            self.conn.touch_port.write('t')
            tmp = self.conn.touch_port.read()
            tmp2 = np.zeros((5, 5))-1
            tmp2[0, 0] = float(tmp)
            yield tmp2

    def stop(self):
        '''
        Disconnect from kinarmdata socket
        '''
        self.conn.touch_port.close()

    def get(self):
        '''
        Get a new kinarm sample
        '''
        return next(self.data)

def make(cls=DataSourceSystem, *args, **kwargs):
    '''
    Docstring
    This ridiculous function dynamically creates a class with a new init function

    Parameters
    ----------

    Returns
    -------
    '''
    def init(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)
    
    dtype = np.dtype((float, (1, )))    
    return type(cls.__name__, (cls,), dict(dtype=dtype, __init__=init))

class TouchDataInterface(object):
    def __init__(self):
        self.touch_port = serial.Serial('/dev/ttyACM2', baudrate=115200)

    def _test_connected(self):
        try:
            self.touch_port.write('t')
            tmp = int(self.touch_port.read())
            self.connected = True
        except:
            print('Error in interfacing w/ touch sensor')
            self.connected = False

    def connect(self):
        self._test_connected()



"""Upload data to google cloud platfom via IoT core. This code is based
on an 'end-to-end' example provided by the GCP documentation"""
import os
os.environ['DJANGO_SETTINGS_MODULE'] = 'db.settings'

import argparse
import datetime
import json
import ssl
import time
import requests

import jwt
import paho.mqtt.client as mqtt


def create_jwt(project_id, private_key_file, algorithm):
    """Create a JWT (https://jwt.io) to establish an MQTT connection.
    Function copied from GCP documentation"""
    token = {
        'iat': datetime.datetime.utcnow(),
        'exp': datetime.datetime.utcnow() + datetime.timedelta(minutes=60),
        'aud': project_id
    }
    with open(private_key_file, 'r') as f:
        private_key = f.read()
    print('Creating JWT using {} from private key file {}'.format(
        algorithm, private_key_file))
    return jwt.encode(token, private_key, algorithm=algorithm)

def error_str(rc):
    """Convert a Paho error to a human readable string."""
    return '{}: {}'.format(rc, mqtt.error_string(rc))


class CloudConfig(object):
    def __init__(self, cloud_info_fname):
        cloud_info = json.load(open(cloud_info_fname))
        for key in cloud_info:
            setattr(self, key, cloud_info[key])

        self.connected = False

def make_client(args):
    # Create the MQTT client and connect to Cloud IoT.
    client = mqtt.Client(
        client_id='projects/{}/locations/{}/registries/{}/devices/{}'.format(
            args.project_id,
            args.cloud_region,
            args.registry_id,
            args.device_id))
    client.username_pw_set(
        username='unused',
        password=create_jwt(
            args.project_id,
            args.private_key_file,
            args.algorithm))
    client.tls_set(ca_certs=args.ca_certs, tls_version=ssl.PROTOCOL_TLSv1_2)

    def on_connect(unused_client, unused_userdata, unused_flags, rc):
        """Callback for when a device connects."""
        print('Connection Result:', error_str(rc))
        args.connected = True

    def on_publish(unused_client, unused_userdata, unused_mid):
        """Callback when the device receives a PUBACK from the MQTT bridge."""
        print('Published message acked.')

    def on_message(unused_client, unused_userdata, message):
        """Callback when the device receives a message on a subscription."""
        payload = str(message.payload.decode('utf-8'))
        data = json.loads(payload)

        if 'action' not in data:
            print('Received message \'{}\' on topic \'{}\' with Qos {}'.format(
                payload, message.topic, str(message.qos)))
            return

        if data['action'] == 'upload_file':
            print("Uploading file {}".format(data['filename']))
            headers = {'Content-type': 'application/octet-stream'}
            r = requests.put(data['url'], data=open(data['full_filename'], 'rb'), headers=headers)
        elif data['action'] == 'check_status':
            print("Checking status of file {}".format(data['filename']))
            import django
            django.setup()
            from . import models
            df = models.DataFile.objects.get(path=data['filename'])
            df.backup_status = data['status'] + ' on {}'.format(str(datetime.datetime.now()))
            # df.archived = data['status'] == 'present'
            df.save()
            print(df.archived, df.backup_status, data)

    client.on_connect = on_connect
    client.on_publish = on_publish
    client.on_message = on_message
    return client

default_cloud_info_fname = os.path.join(os.path.dirname(__file__), '../../config/cloud-config.json')
def upload_json(json_data, cloud_info_fname=default_cloud_info_fname):
    if not os.path.exists(cloud_info_fname):
        print("Cloud upload configuration not found, skipping!")
        return

    args = CloudConfig(cloud_info_fname)
    client = make_client(args)
    client.connect(args.mqtt_bridge_hostname, args.mqtt_bridge_port)
    client.loop_start()

    timeout = 5
    total_time = 0
    while not args.connected and total_time < timeout:
        time.sleep(1)
        total_time += 1

    if not args.connected:
        print('Data not uploaded! Could not connect to MQTT bridge.')

    # This is the topic that the device will publish data to
    mqtt_telemetry_topic = '/devices/{}/events'.format(args.device_id)
    payload = json.dumps(json_data)
    print('Publishing payload', payload)
    client.publish(mqtt_telemetry_topic, payload, qos=1)
    client.disconnect()
    client.loop_stop()
    args.connected = False


def send_message_and_wait(file_data, cloud_info_fname=default_cloud_info_fname):
    """Upload a file by posting a message, retreiving a secret URL, and initiating a resumable upload"""
    if not os.path.exists(cloud_info_fname):
        print("Cloud upload configuration not found, skipping!")
        return

    args = CloudConfig(cloud_info_fname)
    client = make_client(args)
    client.connect(args.mqtt_bridge_hostname, args.mqtt_bridge_port)

    # subscribe to commands topic
    client.subscribe("/devices/{}/commands/#".format(args.device_id), 0)

    client.loop_start()

    timeout = 5
    total_time = 0
    while not args.connected and total_time < timeout:
        time.sleep(1)
        total_time += 1

    if not args.connected:
        print('Data not uploaded! Could not connect to MQTT bridge.')

    # This is the topic that the device will publish data to
    mqtt_telemetry_topic = '/devices/{}/events'.format(args.device_id)
    payload = json.dumps(file_data)
    print('Publishing payload', payload)
    client.publish(mqtt_telemetry_topic, payload, qos=1)

    # wait for signed URL reply
    time.sleep(10)

    client.disconnect()
    client.loop_stop()
    args.connected = False
